"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
ENV Space Util Module

this file is a mypy nightmare due to lots of typing unions and complex types
do not trust mypy, trust the unit tests
"""
import abc
from typing import Annotated

import gymnasium
from pydantic import AfterValidator, BaseModel, ImportString

from corl.libraries.env_space_util import EnvSpaceUtil, convert_gymnasium_space, convert_sample


class SpaceTransformationBase(abc.ABC):
    """
    SpaceTransformationBase provides an anstract class to transform space decfintions
    designed to be used to change space and the boundary of the agent and policy for
    policies that requires specific space types
    """

    def __init__(self, input_space: gymnasium.Space, config: dict) -> None:
        self.config = self.get_validator()(**config)
        self._input_space: gymnasium.Space = input_space
        self._output_space: gymnasium.Space = self.convert_space(self._input_space)

    @abc.abstractmethod
    def convert_space(self, input_space: gymnasium.Space) -> gymnasium.Space:
        """
        Converts input space to transformed space
        """
        raise NotImplementedError

    @abc.abstractmethod
    def convert_sample(self, sample: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type:
        """
        Converts sample from input space to transformed space
        """
        raise NotImplementedError

    @abc.abstractmethod
    def convert_transformed_sample(self, transformed_sample: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type:
        """
        Converts sample from transformed spae back to transformed space
        """
        raise NotImplementedError

    @property
    def input_space(self) -> gymnasium.Space:
        """
        Original input space
        """
        return self._input_space

    @property
    def output_space(self) -> gymnasium.Space:
        """
        Transformed space
        """
        return self._output_space

    @staticmethod
    def get_validator() -> type[BaseModel]:
        """
        Validator for subclass configuration
        """
        return BaseModel


def verify_gym_space(val):
    assert issubclass(val, gymnasium.Space), f"Invalid output_type {val} in gymnasium.Space"
    return val


class SpaceTypeConversionValidator(BaseModel):
    """
    Validator for direct space type conversion
    """

    output_type: Annotated[ImportString, AfterValidator(verify_gym_space)]


class SpaceTypeConversion(SpaceTransformationBase):
    """
    SpaceTypeConversion is a SpaceTransformationBase that attempts to convert
    a space into specified gymnasium.Space types
    """

    def __init__(self, input_space: gymnasium.Space, config: dict) -> None:
        self.config: SpaceTypeConversionValidator
        super().__init__(input_space, config)

    @staticmethod
    def get_validator() -> type[SpaceTypeConversionValidator]:
        """
        SpaceTypeConversionValidator used with SpaceTypeConversion
        """
        return SpaceTypeConversionValidator

    def convert_space(self, input_space: gymnasium.Space) -> gymnasium.Space:
        """
        Converts input space to space defined by output_type
        """
        return convert_gymnasium_space(input_space, self.config.output_type)

    def convert_sample(self, sample: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type:
        """
        Converts sample to space defined by output_type
        """
        return convert_sample(sample=sample, sample_space=self.input_space, output_space=self.output_space)

    def convert_transformed_sample(self, transformed_sample: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type:
        """
        Converts sample in space defined by output_type to the original input space
        """
        return convert_sample(sample=transformed_sample, sample_space=self.output_space, output_space=self.input_space)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Environment Dict Module
"""
import abc
import copy
import logging
import typing
import warnings
from collections import OrderedDict

import numpy as np

from corl.libraries.state_dict import StateDict


class Callback:
    """
    Callback provides basic callback processing for all reward and done functions
    """

    def __init__(self, funcs: typing.Sequence[typing.Callable] | None = None) -> None:
        """
        __init__ constructor

        Parameters
        ----------
        funcs : typing.List[typing.Callable], optional
            List of callale functions, by default None
        """
        self._process_callbacks: list[typing.Callable] = []
        self._logger = logging.getLogger(Callback.__name__)
        if funcs:
            self.register_funcs(funcs)

    def register_func(self, func: typing.Callable) -> None:
        """
        register_func registers a function to the list of valid functions

        Parameters
        ----------
        func : typing.Callable
            The callable function
        """
        # MTB - 10/15/2020 - There seems to be an issue with this when combining multiple reward sets
        # if func is a callable the expression func in self._process_callbacks will always return True
        if isinstance(func, Callback) or func not in self._process_callbacks:
            self._process_callbacks.append(func)
        else:
            warnings.warn("Ignoring a duplicate callback given")

    def register_funcs(self, funcs: typing.Sequence[typing.Callable] | None):
        """
        register_func registers a list of functions to the list of valid functions

        Parameters
        ----------
        func : typing.Callable
            The callable function
        """
        for func in funcs or []:
            self.register_func(func)

    def unregister_func(self, func: typing.Callable) -> None:
        """
        unregister callbacks from processing

        Parameters
        ----------
        key : str
            The callback string to remove
        """
        if func in self._process_callbacks:
            self._process_callbacks.remove(func)

    def unregister_funcs(self) -> None:
        """
        unregister callbacks from processing
        """
        self._process_callbacks.clear()

    def reset_funcs(self):
        """
        reset_funcs func is a callable then attempt to reset its state.

        Parameters
        ----------
        key : str
            The callback string to remove
        """
        for func in self._process_callbacks:
            reset_op = getattr(func, "reset", None)
            if callable(reset_op):
                func.reset()

    @property
    def process_callbacks(self) -> list[typing.Callable]:
        """
        process_callbacks gets the current callbacks

        Returns
        -------
        typing.List[typing.Callable]
            Current list of callbacks
        """
        return self._process_callbacks


class EnvDict(StateDict, Callback):
    """[summary]

    Parameters
    ----------
    StateDict : [type]
        [description]
    Callback : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """

    # list of items to exclude when generating the keys, items, values
    EXCLUDE_KEYS = [
        "_default_kwargs",
        "_reduce_fn",
        "_reduce_fn_kwargs",
        "_processing_funcs",
        "_recurse",
    ]

    class DuplicateName(RuntimeError):
        """Exception class for callbacks with duplicate names"""

    def __init__(
        self,
        processing_funcs: typing.Sequence[typing.Callable] | None = None,
        reduce_fn: typing.Callable | None = None,
        reduce_fn_kwargs=None,
        **kwargs,
    ) -> None:
        """
        __init__ environment dictionary constructor

        Parameters
        ----------
        processing_funcs : typing.List[typing.Callable], optional
            List of functions to call by the environment, by default None
        reduce_fn : typing.Callable, optional
            function used to reduce the results, by default None
        reduce_fn_kwargs : [type], optional
            [description], by default None
        """
        self._default_kwargs = None
        self._reduce_fn = reduce_fn
        self._reduce_fn_kwargs = reduce_fn_kwargs or {}
        self._set_default_kwargs(kwargs)
        StateDict.__init__(self, **kwargs)
        Callback.__init__(self, processing_funcs)

    def __call__(self, *args, **kwargs) -> tuple[OrderedDict, OrderedDict]:
        """
        __call__ Callable function for the environment dictionary type

        Returns
        -------
        typing.Tuple[OrderedDict, OrderedDict]
            The reduced rewards and theret information
        """
        r = [
            self._default_kwargs,
        ]
        ret_info: OrderedDict = OrderedDict()

        def merge(source, destination):
            """
            run me with nosetests --with-doctest file.py

            >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
            >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
            >>> merge(b, a) == { 'first' : { 'all_rows' : { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }
            True
            """
            for key, value in source.items():
                if isinstance(value, dict):
                    # get node or create one
                    node = destination.setdefault(key, {})
                    merge(value, node)
                else:
                    destination[key] = value

            return destination

        for func in self._filtered_process_callbacks:
            if isinstance(func, OrderedDict):
                ret = func(*args, **kwargs)
                rew, info = ret
                r.append(rew)
                ret_info = merge(ret_info, info)

        for func in self._filtered_process_callbacks:
            if not isinstance(func, OrderedDict):
                ret = func(*args, **kwargs)
                # single value
                r.append(copy.deepcopy(ret))

                try:
                    name = func.__name__
                except:  # noqa: E722
                    name = func.name  # type: ignore

                # This only affects the info dictionary that is returned.  As the code below merges the output for all agents together,
                # the __all__ entry would be overwritten to only provide the information of the last agent, which could be confusing or
                # inaccurate.  Therefore, remove __all__ from the returned information.
                if "__all__" in ret:
                    del ret["__all__"]

                if name in ret_info:
                    common_keys = set(ret.keys()) & set(ret_info[name].keys())
                    if common_keys:
                        raise self.DuplicateName(f"{name} has common keys: {common_keys}")
                    ret_info[name].update(**ret)
                else:
                    ret_info[name] = ret

        # TODO link with reduce and ret info
        return self._reduce(r, **self._reduce_fn_kwargs), ret_info  # type: ignore

    @property
    def _filtered_process_callbacks(self) -> list[typing.Callable]:
        """Set of callbacks that have been filtered by subclass logic.

        Default implementation is all callbacks

        Returns
        -------
        typing.List[typing.Callable]
            Callbacks to apply
        """
        return self._process_callbacks

    @abc.abstractmethod
    def _reduce(self, r: typing.Callable, **kwargs):
        """
        _reduce user defined reduce function for processing

        Parameters
        ----------
        r : typing.Callable
            The reduce function to use
        """

    def _set_default_kwargs(self, kwargs):
        """
        _set_default_kwargs [summary]

        [extended_summary]

        Parameters
        ----------
        kwargs : [type]
            [description]
        """
        if self._default_kwargs is None:
            self._default_kwargs = copy.deepcopy(kwargs)

    def reset(self):
        """
        reset [summary]
        """
        # EnvDict.__init__(self, **self._default_kwargs)
        self.reset_funcs()

    def _filtered_self(self):
        tmp = {k: v for k, v in super().items() if k not in self.EXCLUDE_KEYS}
        return StateDict(tmp)

    def keys(self):
        return self._filtered_self().keys()

    def values(self):
        return self._filtered_self().values()

    def items(self):
        return self._filtered_self().items()

    def to_dict(self):
        return self._filtered_self().to_dict()

    @property
    def name(self):
        """[summary]

        Returns
        -------
        [type]
            [description]
        """
        return type(self).__name__

    @staticmethod
    def observation_deep_copy(d1, d2):
        """[summary]

        Parameters
        ----------
        d1 : [type]
            [description]
        d2 : [type]
            [description]
        """
        # d1 = pickle.loads(pickle.dumps(d2))
        for (k1, v1), (k2, v2) in zip(d1.items(), d2.items()):
            if isinstance(v1, dict):
                EnvDict.observation_deep_copy(v1, v2)
            else:
                d1[k1] = copy.deepcopy(d2[k2])


class DoneDict(EnvDict):
    """[summary]

    Parameters
    ----------
    EnvDict : [type]
        [description]
    """

    def __init__(
        self,
        processing_funcs: list[typing.Callable] | None = None,
        reduce_fn: typing.Callable | None = None,
        reduce_fn_kwargs=None,
        **kwargs,
    ) -> None:
        super().__init__(processing_funcs=processing_funcs, reduce_fn=reduce_fn, reduce_fn_kwargs=reduce_fn_kwargs, **kwargs)

        self._agent_filter: typing.Iterable[str] | None = None

    def __call__(self, *args, **kwargs) -> tuple[OrderedDict[typing.Any, typing.Any], OrderedDict[typing.Any, typing.Any]]:
        """
        __call__ Callable function for the done dictionary type

        Returns
        -------
        typing.Tuple[OrderedDict, OrderedDict]
            The done information
        """
        r = super().__call__(*args, **kwargs)

        # Check for bool type in return value
        for key, value in r[0].items():
            if isinstance(value, np.bool_):
                r[0][key] = bool(value)
            elif not isinstance(value, bool):
                raise TypeError(f"DoneDict __call__ return is not type bool for key: {key}")
        # remap from platform: {Done: {platform: bool}} -> platform: {Done: bool}
        tmp: OrderedDict[str, OrderedDict[str, bool]] = OrderedDict(
            [(platform_name, OrderedDict()) for platform_name in r[0] if platform_name != "__all__"]
        )
        for key0, value0 in r[1].items():
            for platform_name, platform_info in tmp.items():
                platform_info[key0] = bool(value0.get(platform_name, False))

        return (r[0], tmp)

    @property
    def _filtered_process_callbacks(self) -> list[typing.Callable]:
        if self._agent_filter is None:
            return super()._filtered_process_callbacks

        # Avoid circular import
        from corl.dones.done_func_base import DoneFuncBase

        return [x for x in self._process_callbacks if not isinstance(x, DoneFuncBase) or x.agent in self._agent_filter]

    def set_alive_agents(self, alive_agents: typing.Iterable[str]) -> None:
        """Specify which agents are alive

        This is used to determine which callbacks to call.

        Parameters
        ----------
        alive_agents : typing.Iterable[str]
            Agents that are currently alive.
        """
        self._agent_filter = alive_agents

    def _reduce(self, r, **kwargs):
        self._reduce_fn = self._reduce_fn or np.any
        tmp = StateDict.stack_values(r)
        tmp = {k: self._reduce_fn(v, **kwargs) for k, v in tmp.items()}
        return StateDict(sorted(tmp.items()))


class InfoDict(EnvDict):
    """[summary]

    Parameters
    ----------
    EnvDict : [type]
        [description]
    """

    def _reduce(self, r, **kwargs):
        self._reduce_fn = self._reduce_fn or (lambda x: {})
        tmp = StateDict.stack_values(r)
        tmp = {k: self._reduce_fn(v, **kwargs) for k, v in tmp.items()}
        return StateDict(sorted(tmp.items()), recurse=False)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

AvailablePlatforms
"""
import logging
import typing
from collections import OrderedDict

import gymnasium
import numpy as np
from pydantic import field_validator

from corl.glues.base_dict_wrapper import BaseDictWrapperGlue, BaseDictWrapperGlueValidator
from corl.glues.common.controller_glue import ControllerGlue
from corl.glues.common.observe_sensor import ObserveSensor
from corl.glues.controller_wrappers.obs_relative_delta_controller import RelativeObsDeltaAction
from corl.libraries.env_space_util import EnvSpaceUtil


class RelativeObsDeltaActionDictValidator(BaseDictWrapperGlueValidator):
    """
    step_size:      A dict that contains a floating point scalar for each action in the action space,
                    by which the corresponding delta action is scaled prior to converting the action
                    to the wrapped space.
                    e.g. A throttle DeltaAction.apply_action([0.2]) with step_size=[.05] would move the
                    absolute throttle position to 0.01 higher than it was at the end of the last step.
    """

    step_size: float = 1.0
    obs_index: int | None = 0
    is_wrap: bool = False
    initial_value: float | None = None

    @field_validator("step_size")
    @classmethod
    def check_step_scale(cls, v):
        """
        verifies range of step scale values
        """
        if v >= 1.0 or v < 0:
            raise ValueError("RelativeObsDeltaActionValidator got step size of more that 1.0 or less than 0")
        return v


class RelativeObsDeltaActionDict(BaseDictWrapperGlue, RelativeObsDeltaAction):  # type: ignore[misc]
    """
    RelativeObsDeltaActionDict is a glue class that wraps another glue class.
    It treats the actions passed to it as a delta from a linked observation
    E.G. if the wrapped action space has has roll as one of the controls, then a delta action of
    0.2 would move the absolute roll position 0.2 higher than it is as measured by the linked roll sensor.
    """

    def __init__(self, **kwargs) -> None:
        self.config: RelativeObsDeltaActionDictValidator  # type: ignore[assignment]
        BaseDictWrapperGlue.__init__(**kwargs)
        self._logger = logging.getLogger(RelativeObsDeltaActionDict.__name__)

        controller_keys = self.glues().keys()
        if "controller" not in controller_keys:
            raise KeyError("Missing key: controller")
        if "sensor" not in controller_keys:
            raise KeyError("Missing key: sensor")

        self.controller: ControllerGlue = typing.cast(ControllerGlue, self.glues()["controller"])
        if not isinstance(self.controller, ControllerGlue):
            raise RuntimeError(
                "Error: RelativeObsDeltaActionDict expects the glue wrapped on the 'controller' key to be a ControllerGlue, "
                f"got {self.controller}"
            )
        self.relative_obs_glue: ObserveSensor = typing.cast(ObserveSensor, self.glues()["sensor"])
        if not isinstance(self.relative_obs_glue, ObserveSensor):
            raise RuntimeError(
                "Error: RelativeObsDeltaActionDict expects the glue wrapped on the 'sensor' key to be a ObserveSensor, "
                f"got {self.relative_obs_glue}"
            )

        # verify that the config setup is not going to get the user into a situation where they are
        # only accessing one part of the obs but applying that obs as the base position for multiple actions

        tmp = self.controller.action_space
        if not isinstance(tmp, gymnasium.spaces.Dict):
            raise RuntimeError("obs relative delta controller only knows how to operate on dictionary action spaces currently")
        if self.config.obs_index and len(next(iter(tmp.spaces.values())).low) != 1:  # type: ignore
            raise RuntimeError(
                f"ERROR: your glue {self.get_unique_name()} has an action space length of more than 1, "
                "but you specified though obs_index to access only 1 component of the obs "
                "from the wrapped observe Sensor, to fix this error in your config for this glue define 'obs_index': null"
            )

        self.step_size = EnvSpaceUtil.convert_config_param_to_space(action_space=tmp, parameter=self.config.step_size)

        self._is_wrap = self.config.is_wrap

        self.saved_action_deltas = OrderedDict()
        assert isinstance(self.action_space, gymnasium.spaces.Dict)
        for space_name, space in self.action_space.items():
            if self.config.initial_value is not None:
                self.saved_action_deltas[space_name] = np.asarray([self.config.initial_value], dtype=np.float32)
            else:
                assert isinstance(space, gymnasium.spaces.Box)
                self.saved_action_deltas[space_name] = space.low

    @staticmethod
    def get_validator() -> type[RelativeObsDeltaActionDictValidator]:  # type: ignore[override]
        return RelativeObsDeltaActionDictValidator



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Callbacks to do logging of episodes
"""
from ray.rllib import BaseEnv, RolloutWorker
from ray.rllib.algorithms.callbacks import DefaultCallbacks
from ray.rllib.evaluation.episode import Episode
from ray.rllib.evaluation.episode_v2 import EpisodeV2
from ray.rllib.policy import Policy
from ray.rllib.utils.typing import PolicyID


class EpisodeInterrupted(Exception):
    def __init__(self):
        super().__init__("EpisodeInterrupted")


class InterruptableCallback(DefaultCallbacks):
    def __init__(self, legacy_callbacks_dict: dict[str, callable] | None = None):  # type: ignore
        super().__init__(legacy_callbacks_dict)  # type: ignore

        self._reset_required = False

    # def on_episode_start(
    #     self,
    #     *,
    #     worker: RolloutWorker,
    #     base_env: BaseEnv,
    #     policies: dict[PolicyID, Policy],
    #     episode: Episode | EpisodeV2,
    #     env_index: int | None = None,
    #     **kwargs,
    # ) -> None:
    #     self._reset_required = False

    def on_episode_step(
        self,
        *,
        worker: RolloutWorker,
        base_env: BaseEnv,
        policies: dict[PolicyID, Policy] | None = None,
        episode: Episode | EpisodeV2,
        env_index: int | None = None,
        **kwargs,
    ) -> None:
        if self._reset_required:
            self._reset_required = False
            raise EpisodeInterrupted

    def stop(self):
        self._reset_required = True



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.evaluation.metrics.generator import MetricGeneratorTerminalEventScope
from corl.evaluation.metrics.metric import Metric
from corl.evaluation.metrics.types.terminals.real import Real


class EpisodeLength_Steps(MetricGeneratorTerminalEventScope):
    """Real number indicating seconds the event took

    Metric: Real
    Scope: Event
    """

    def generate_metric(self, params: EpisodeArtifact, **kwargs) -> Metric:  # noqa: PLR6301
        return Real(len(params.steps))



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from collections import OrderedDict


class ObsBuffer:
    """Wrapper for OBS functions"""

    def __init__(self) -> None:
        """Constructor"""
        self._buffer1: OrderedDict = OrderedDict()
        self._buffer2: OrderedDict = OrderedDict()
        self._index = 0

    def update_obs_pointer(self):
        """get the next observation pointer for next and current"""
        self._index += 1

    @property
    def observation(self) -> OrderedDict:
        """Every other call we update the pointer to the correct obs buffer
        Returns
        -------
        OrderedDict
            The observation dictionary
        """
        return self._buffer1 if self._index % 2 == 0 else self._buffer2

    @observation.setter
    def observation(self, data: OrderedDict) -> None:
        """sets the observation

        Parameters
        ----------
        data : OrderedDict
            The update data
        """
        if self._index % 2 == 0:
            # Environment.ObsBuffer.update_dict(self._buffer1, data)
            self._buffer1 = data
        else:
            # Environment.ObsBuffer.update_dict(self._buffer2, data)
            self._buffer2 = data

    @property
    def next_observation(self) -> OrderedDict:
        """Every other call we update the pointer to the correct next obs buffer
        Returns
        -------
        OrderedDict
            The next observation dictionary
        """
        return self._buffer2 if self._index % 2 == 0 else self._buffer1

    @next_observation.setter
    def next_observation(self, data: OrderedDict) -> None:
        """sets the next observation

        Parameters
        ----------
        data : OrderedDict
            The update data
        """
        if self._index % 2 == 0:
            # Environment.ObsBuffer.update_dict(self._buffer2, data)
            self._buffer2 = data
        else:
            # Environment.ObsBuffer.update_dict(self._buffer1, data)
            self._buffer1 = data



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import abc
from collections import OrderedDict

import gymnasium
from pydantic import ConfigDict

from corl.dones.done_func_base import DoneFuncBase, DoneFuncBaseValidator
from corl.simulators.base_simulator import BaseSimulatorState


class BaseMultiWrapperDoneValidator(DoneFuncBaseValidator):
    """
    wrapped - the wrapped done instance
    """

    wrapped: list[DoneFuncBase]
    model_config = ConfigDict(arbitrary_types_allowed=True)


class BaseMultiWrapperDone(DoneFuncBase):
    """A base object that dones can inherit in order to "wrap" a single done instance"""

    def __init__(self, **kwargs) -> None:
        self.config: BaseMultiWrapperDoneValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[BaseMultiWrapperDoneValidator]:
        return BaseMultiWrapperDoneValidator

    def dones(self) -> list[DoneFuncBase]:
        """Get the wrapped done instance"""
        return self.config.wrapped

    @abc.abstractmethod
    def __call__(
        self,
        observation: OrderedDict,
        action: OrderedDict,
        next_observation: OrderedDict,
        next_state: BaseSimulatorState,
        observation_space: gymnasium.Space,
        observation_units: OrderedDict,
    ) -> bool:
        ...



"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Functions for the streamlit episode playback
"""

import plotly.graph_objs as go
import stqdm
import streamlit as st
from streamlit.runtime.state.session_state_proxy import SessionStateProxy

from corl.libraries.functor import Functor
from corl.visualization.plotly_animator import FrameAxisBounds
from corl.visualization.streamlit_app.utils import center_plotly_figure, checkbox_grid, use_container_width


@st.cache_resource
def display_animation(fig: go.Figure, test_case: int):
    """Displays the plotly chart"""
    st.markdown(f"#### Showing Animation for Test Case: {test_case}")
    center_plotly_figure(fig)


def trajectory_playback_section(
    animator: Functor, sess_state: SessionStateProxy, episode_artifacts_dict, frame_axis_bounds: FrameAxisBounds | None = None
) -> SessionStateProxy:
    """Trajectory playback section for streamlit

    Parameters
    ----------
    animator: Functor
        A Functor defining the animator
    sess_state : SessionStateProxy
        The streamlit session state
    episode_artifacts_dict : dict
        A dictionary containing the episode artifacts, where the key is the test
        case number and the value is the episode artifact.
    frame_axis_bounds: FrameAxisBounds, optional
        Sets the default frame axis bounds for the animation.

    Returns
    -------
    SessionStateProxy
        The streamlit session state
    """

    st.markdown("#### Trajectory Animation")
    st.markdown("**The test case selection control is in the `AgentSteps` section**")
    default_x_bounds = frame_axis_bounds.xbounds if frame_axis_bounds else (0, 100)
    default_y_bounds = frame_axis_bounds.ybounds if frame_axis_bounds else (0, 100)
    default_z_bounds = frame_axis_bounds.zbounds if frame_axis_bounds and frame_axis_bounds.zbounds else (0, 100)
    animation_cls = animator.functor(
        episode_artifact=episode_artifacts_dict[sess_state.test_case_selected], progress_bar_func=stqdm.stqdm, **animator.config
    )
    if animation_cls.get_deserializer() is None:
        platform_serializer = str(episode_artifacts_dict[sess_state.test_case_selected].platform_serializer)
        st.warning(f"Skipping Trajectory Animation. No platform deserializer found for {platform_serializer}")
        return sess_state

    # #####################################################################
    # When the frame axis bounds is not provided, display a section
    # for user input for the frame axis bounds.
    # #####################################################################
    st.markdown(
        """
        Set the frame axis bounds for the animation. The animation will
        be played back with these bounds.
        """
    )
    columns = st.columns(6)
    dimensions = animation_cls.get_dimensionality()
    with columns[0]:
        min_x = st.number_input(label="x-axis min", value=default_x_bounds[0])
    with columns[1]:
        max_x = st.number_input(label="x-axis max", value=default_x_bounds[1])
    with columns[2]:
        min_y = st.number_input(label="y-axis min", value=default_y_bounds[0])
    with columns[3]:
        max_y = st.number_input(label="y-axis max", value=default_y_bounds[1])
    frame_axis_obj = FrameAxisBounds(xbounds=(min_x, max_x), ybounds=(min_y, max_y), zbounds=None)

    if dimensions == 3:
        with columns[4]:
            min_z = st.number_input(label="z-axis min", value=default_z_bounds[0])
            max_z = st.number_input(label="z-axis max", value=default_z_bounds[1])
        frame_axis_bounds = FrameAxisBounds(xbounds=(min_x, max_x), ybounds=(min_y, max_y), zbounds=(min_z, max_z))

    # #####################################################################
    # Selection grid for metadata to display above platforms
    # #####################################################################
    possible_selections = animation_cls.get_metadata_names()
    selections = None
    if possible_selections:
        with st.expander(label="**Select Attributes to Display Above Platforms**", expanded=True):
            selections = checkbox_grid(possible_choices=possible_selections, key_prefix="traj_anim")
    animation_button = st.button("Generate Animation")

    # When the test case changes, do not display the animation
    if (
        hasattr(sess_state, "test_case_for_animation")
        and hasattr(sess_state, "test_case_selected")
        and sess_state.test_case_for_animation != sess_state.test_case_selected
    ):
        sess_state.generated_animation = False

    if animation_button:
        animation_cls.set_frame_axis_bounds(frame_axis_obj)
        sess_state.traject_fig, sess_state.traj_stored_frames, sess_state.traj_tables = animation_cls.generate_animation(selections)
        sess_state.generated_animation = True
        sess_state.test_case_for_animation = sess_state.test_case_selected

    if animation_button or sess_state.get("generated_animation"):
        st.markdown("**Note**: Toggle tabs to see underlying raw data in tabular format per platform.")

        platforms = list(sess_state.traj_tables.keys())
        tabs = st.tabs(["Animation"] + [f"Data Table: {platform}" for platform in platforms])
        with tabs[0]:
            display_animation(sess_state.traject_fig, sess_state.test_case_for_animation)
        for idx, platform in enumerate(platforms):
            with tabs[idx + 1]:
                st.markdown(f"##### Underlying Steps Data: {platform}")
                st.dataframe(sess_state.traj_tables[platform], use_container_width=use_container_width)
    return sess_state



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Base Agent Class
"""
from __future__ import annotations

import abc
import collections
import copy
import fractions
import typing
from collections.abc import MutableMapping
from functools import cached_property
from graphlib import TopologicalSorter
from typing import Any, Literal

import flatten_dict
import gymnasium
from pydantic import BaseModel, ConfigDict, Field, ImportString, field_validator
from tree import map_structure

from corl.dones.done_func_base import DoneFuncBase
from corl.dones.episode_length_done import EpisodeLengthDone
from corl.episode_parameter_providers import EpisodeParameterProvider, Randomness
from corl.glues.base_glue import BaseAgentGlue, TrainingExportBehavior
from corl.libraries.env_space_util import EnvSpaceUtil
from corl.libraries.factory import Factory
from corl.libraries.functor import Functor, FunctorDictWrapper, FunctorMultiWrapper, FunctorWrapper, ObjectStoreElem
from corl.libraries.parameters import Parameter
from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.property import DictProp, NestedDict, Prop
from corl.libraries.space_transformations import SpaceTransformationBase
from corl.libraries.units import corl_get_ureg
from corl.rewards.reward_func_base import RewardFuncBase
from corl.simulators.base_simulator import validation_helper_units_and_parameters


class PartsFunctor(BaseModel):
    """
    - part: The name of the part. This should be registered to a corresponding platform part class in the plugin library.
    - config: The specific configuration dictionary expected by the platform part.
    """

    part: str
    config: dict[str, Any] = {}
    references: dict[str, str] = {}


class AgentParseBase(BaseModel):
    """
    - agent: The BaseAgent class representing the agent.
    - config: The agent specific configuration including glues, reward, dones, parts, etc.
    """

    agent: ImportString
    config: dict[str, Any]

    @field_validator("agent")
    @classmethod
    def check_agent(cls, v):
        """Check if agent subclass AgentBase"""
        if not issubclass(v, BaseAgent):
            raise ValueError(f"Agents must subclass BaseAgent, but is of type {v}")
        return v


class AgentParseInfo(BaseModel):
    """
    - class_config: The agent class and its configuration
    - platform_name: The name of the platform
    - policy_config: Configuration of the policy of this agent
    """

    class_config: AgentParseBase
    platform_names: list[str]
    policy_config: dict[str, Any]


class PlatformParseInfo(BaseModel):
    """
    - platform_config
    """

    platform_config: dict[str, Any]


class BaseAgentEppParameters(BaseModel):
    """
    glues: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
        keys: glue name, parameter name

    rewards: typing.Dict[str, typing.Dict[str, typing.Dict[str, typing.Any]]] = {}
        keys: reward name, ? name, parameter name

    dones: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
        keys: done name, parameter name

    reference_store: typing.Dict[str, typing.Any] = {}
        keys: reference name

    simulator_reset: typing.Dict[str, typing.Any] = {}
        keys: whatever the simulator wants, but it needs to be kwargs to simulator reset
    """

    glues: dict[str, dict[str, typing.Any]] = {}
    rewards: dict[str, dict[str, typing.Any]] = {}
    dones: dict[str, dict[str, typing.Any]] = {}
    reference_store: dict[str, typing.Any] = {}
    simulator_reset: dict[str, typing.Any] = {}

    @staticmethod
    def _validate_leaves_are_parameters(obj):
        if isinstance(obj, dict):
            for value in obj.values():
                BaseAgentEppParameters._validate_leaves_are_parameters(value)
        elif not isinstance(obj, Parameter):
            raise TypeError(f"Invalid type: {type(obj)} (required type: {Parameter.__qualname__})")

    @field_validator("glues", "rewards", "dones", "reference_store", "simulator_reset")
    @classmethod
    def validate_leaves_are_parameters(cls, v):
        """
        Verifies the outer most leaf nodes of a config are parameter types
        """
        BaseAgentEppParameters._validate_leaves_are_parameters(v)
        return v


class AgentSpaceTransformation(BaseModel):
    """
    AgentSpaceTransformation

        transformation: Type[SpaceTransformationBase]
        config: Dict
    """

    transformation: ImportString
    config: dict = {}

    @field_validator("transformation")
    @classmethod
    def validate_transformation(cls, val):
        """
        validates transformation subclass type
        """
        if issubclass(val, SpaceTransformationBase):
            return val

        raise TypeError(f"Invalid output_type {type(val)} in SpaceTypeConversionValidator")


def check_rewards(v):
    """Check if rewards subclass RewardFuncBase"""
    if not issubclass(v.functor, RewardFuncBase):
        raise TypeError(f"Reward functors must subclass RewardFuncBase, but reward {v.name} is of type {v.functor}")
    return v


def check_dones(v):
    """Check if dones subclass DoneFuncBase"""
    if not issubclass(v.functor, DoneFuncBase):
        raise TypeError(f"Done functors must subclass DoneFuncBase, but done {v.name} is of type {v.functor}")
    if issubclass(v.functor, EpisodeLengthDone):
        raise ValueError("Cannot specify EpisodeLengthDone as it is automatically added")
    return v


class BaseAgentParser(BaseModel):
    """
    - parts: A list of part configurations.
    - glues: A list of glue functor configurations.
    - rewards: A optional list of reward functor configurations
    - dones: An optional list of done functor configurations
    """

    # these can be sensors, or controller
    agent_name: str
    parts: list[list[PartsFunctor]]  # this uses plugin loader to load
    reference_store: dict[str, ObjectStoreElem] = {}
    glues: list[list[FunctorMultiWrapper | (FunctorWrapper | (Functor | FunctorDictWrapper))]]
    rewards: list[FunctorMultiWrapper | (FunctorWrapper | (Functor | FunctorDictWrapper))] = []
    dones: list[FunctorMultiWrapper | (FunctorWrapper | (Functor | FunctorDictWrapper))] = Field(default=[], validate_default=True)
    simulator_reset_parameters: dict[str, Any] = {}
    # frame_rate in Hz (i.e. .25 indicates this agent will process when sim_tim % 4 == 0)
    frame_rate: float = 1.0
    platform_names: list[str]
    multiple_workers: bool = False
    policy_action_transformation: AgentSpaceTransformation | None = None
    policy_observation_transformation: AgentSpaceTransformation | None = None
    # modifies the normalization for any glue created by this agent
    normalization_master_switch: Literal["ALL_ON", "DEFAULT", "ALL_OFF"] = "DEFAULT"
    episode_parameter_provider: Factory
    episode_parameter_provider_parameters: BaseAgentEppParameters = Field(validate_default=True, default=None)  # type: ignore
    epp: EpisodeParameterProvider = Field(validate_default=True, default=None)  # type: ignore
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("parts", "glues", mode="before")
    @classmethod
    def convert_parts_and_glues(cls, v):
        """Check if parts subclass BaseAgentGlue"""
        if isinstance(v, list):
            if all(isinstance(elem, list) for elem in v):
                return v
            return [v]
        raise TypeError(f"Parts and glues must be List not {type(v)}")

    @field_validator("glues")
    def check_glues(cls, v, info):
        """Check if glues subclass BaseAgentGlue"""
        for glue_set in v:
            for glue in glue_set:
                if not issubclass(glue.functor, BaseAgentGlue):
                    raise TypeError(f"Glue functors must subclass BaseAgentGlue, but glue {v.name} is of type {v.functor}")

        assert "agent_name" in info.data
        assert "parts" in info.data

        for glues, parts in zip(v, info.data["parts"]):
            for glue in glues:
                cls._check_glue_parts(glue, parts, info.data)

        return v

    @classmethod
    def _check_glue_parts(cls, v, parts, values) -> None:
        for part_type in ["sensor", "controller"]:
            if part_type in v.config:
                assert any(
                    x.config.get("name", x.part) == v.config[part_type] for x in parts
                ), f'{v.config[part_type]} not found in the parts of {values["agent_name"]}'

        if isinstance(v, FunctorWrapper):
            cls._check_glue_parts(v.wrapped, parts, values)
        elif isinstance(v, FunctorMultiWrapper):
            for elem in v.wrapped:
                cls._check_glue_parts(elem, parts, values)
        elif isinstance(v, FunctorDictWrapper):
            for elem in v.wrapped.values():
                cls._check_glue_parts(elem, parts, values)

    @field_validator("simulator_reset_parameters", mode="before")
    @classmethod
    def update_units_and_parameters(cls, v):
        """Update simulation reset parameters to meet base simulator requirements."""
        return validation_helper_units_and_parameters(v)

    @field_validator("episode_parameter_provider")
    def check_epp(cls, epp_factory, values):
        """Check if episode parameter provider subclass EpisodeParameterProvider"""
        if not issubclass(epp_factory.type, EpisodeParameterProvider):
            raise TypeError(f"Episode parameter providers must subclass EpisodeParameterProvider, but is is of type {epp_factory.type}")

        # replace %%AGENT%% with agent_name
        epp_params = flatten_dict.flatten(epp_factory.config)
        for key, value in epp_params.items():
            if isinstance(value, str):
                epp_params[key] = value.replace(r"%%AGENT%%", values.data["agent_name"])
        epp_factory.config = flatten_dict.unflatten(epp_params)

        return epp_factory

    @field_validator("episode_parameter_provider_parameters", mode="before")
    def build_episode_parameter_provider_parameters(cls, _v, values) -> BaseAgentEppParameters:
        """Create the episode parameter provider config"""

        for key in ["reference_store", "glues", "rewards", "dones", "simulator_reset_parameters"]:
            assert key in values.data

        reference_parameters: dict[str, Parameter] = {
            ref_name: ref_value for ref_name, ref_value in values.data["reference_store"].items() if isinstance(ref_value, Parameter)
        }
        glue_parameters: dict[str, dict[str, Parameter]] = {}
        for glue_set in values.data["glues"]:
            for functor in glue_set:
                functor.add_to_parameter_store(glue_parameters)

        reward_parameters: dict[str, dict[str, Parameter]] = {}
        for functor in values.data["rewards"]:
            functor.add_to_parameter_store(reward_parameters)

        done_parameters: dict[str, dict[str, Parameter]] = {}
        for functor in values.data["dones"]:
            functor.add_to_parameter_store(done_parameters)

        flat_data = flatten_dict.flatten(values.data["simulator_reset_parameters"])
        for key, value in flat_data.items():
            if isinstance(value, BaseModel):
                flat_data[key] = value.dict()
        expanded_sim_reset_params = flatten_dict.unflatten(flat_data)

        sim_parameters_flat = {
            name: param for name, param in flatten_dict.flatten(expanded_sim_reset_params).items() if isinstance(param, Parameter)
        }
        sim_parameters = flatten_dict.unflatten(sim_parameters_flat)

        return BaseAgentEppParameters(
            glues=glue_parameters,
            rewards=reward_parameters,
            dones=done_parameters,
            reference_store=reference_parameters,
            simulator_reset=sim_parameters,
        )

    @field_validator("epp", mode="before")
    def build_epp(cls, epp, values):
        """Builds an instance of an EpisodeParameterProvider if necessary"""
        if epp is None:
            assert "episode_parameter_provider_parameters" in values.data

            epp_params = dict(values.data["episode_parameter_provider_parameters"])
            flat_epp_params = flatten_dict.flatten(epp_params)
            epp = values.data["episode_parameter_provider"].build(parameters=flat_epp_params)
        return epp

    @field_validator("frame_rate")
    def simplify_frame_rate(cls, v):
        """Expand the precision of the frame rate (e.g. .333 -> .3333333333334)"""
        f = fractions.Fraction(v).limit_denominator(20)
        return f.numerator / f.denominator


class BaseAgent:  # noqa: PLR0904
    """
    Base class representing an agent in an environment.
    """

    def __init__(self, **kwargs) -> None:
        self.config: BaseAgentParser = self.get_validator()(**kwargs)
        self.agent_glue_dict: dict[str, BaseAgentGlue] = {}
        self.agent_reward_functors: list[RewardFuncBase] = []
        self.agent_glue_order: tuple = ()

        self._action_space_transformation: SpaceTransformationBase | None = None
        self._obs_space_transformation: SpaceTransformationBase | None = None

        # self._agent_glue_obs_export_behavior = {}

        # Sample parameter provider
        # This RNG only used here.  Normal use uses the one from the environment.
        rng, _ = gymnasium.utils.seeding.np_random(0)
        self.fill_parameters(rng=rng, default_parameters=True)
        assert len(self.config.platform_names) == 1, "BaseAgent can only have one platform assigned to it currently"

    @property
    def platform_names(self) -> list[str]:
        """
        Returns the name of the platform this agent is attached to

        Returns
        -------
        str:
            The name of the platform
        """
        return self.config.platform_names

    @property
    def trainable(self) -> bool:
        """
        Flag denoting if agent is trainable.

        Returns
        -------
        bool:
            False
        """
        return False

    @staticmethod
    def get_validator() -> type[BaseAgentParser]:
        """
        Get the validator used to validate the kwargs passed to BaseAgent.

        Returns
        -------
        BaseAgentParser
            A BaseAgent kwargs parser and validator.
        """
        return BaseAgentParser

    @property
    def frame_rate(self) -> float:
        """Get the frame rate this agent runs at"""
        return self.config.frame_rate

    def fill_parameters(self, rng: Randomness, env_epp_ctx: dict | None = None, default_parameters: bool = False) -> None:
        """Sample the episode parameter provider to fill the local variable store."""
        if default_parameters:
            current_parameters = self.config.epp.config.parameters
        else:
            current_parameters, _, _ = self.config.epp.get_params(rng, env_epp_ctx)

        flat_variable_store = {}
        for k, v in current_parameters.items():
            param_value = v.get_value(rng, {})
            if isinstance(param_value, dict):
                for param_k, param_v in param_value.items():
                    # for the dict case the param has a path that would be something like
                    # ('simulator_reset', 'platforms', 'paddle0', 'y_and_vel')
                    # we want to collapse down to the nested params so
                    # ('simulator_reset', 'platforms', 'paddle0', 'y')
                    # and ('simulator_reset', 'platforms', 'paddle0', 'vel')
                    # this may be evil and need to be reworked, but not 100% sure
                    flat_variable_store[(*k[:-1], param_k)] = param_v
            else:
                flat_variable_store[k] = param_value

        self.local_variable_store = flatten_dict.unflatten(flat_variable_store)

    def get_simulator_reset_parameters(self) -> dict[str, typing.Any]:
        """Return the local parameters needed within the simulator reset"""
        output = copy.deepcopy(self.config.simulator_reset_parameters)

        def expand_and_update(data: MutableMapping[str, Any], simulator_reset: typing.Mapping[str, Any]):
            for key, value in data.items():
                if isinstance(value, BaseModel):
                    expand_and_update(vars(value), simulator_reset.get(key, {}))
                elif isinstance(value, MutableMapping):
                    expand_and_update(value, simulator_reset.get(key, {}))
                elif key in simulator_reset:
                    data[key] = simulator_reset[key]

        expand_and_update(output, self.local_variable_store.get("simulator_reset", {}))

        return output

    def get_platform_parts(self, simulator, platform_dict) -> dict[str, list]:
        """
        Gets a list of the agent's platform parts.

        Parameters
        ----------
        simulator: Simulator class used by environment to simulate agent actions.
        platform_type: A mapping of platform_names to platform type enumeration corresponding to the agent's platform

        Returns
        -------
        list:
            List of platform parts
        """
        my_parts: dict[str, list] = {platform_name: [] for platform_name in self.config.platform_names}
        for platform_name, platform_parts in zip(self.config.platform_names, self.config.parts):
            mapped_platform = platform_dict[platform_name]
            for part in platform_parts:
                tmp = PluginLibrary.FindMatch(part.part, {"simulator": simulator, "platform_type": mapped_platform})
                for ref_name, ref_key in part.references.items():
                    if ref_key not in self.config.reference_store:
                        raise RuntimeError(f"Part reference {ref_key} must be in the agent reference store.")
                    ref_value = self.config.reference_store[ref_key]
                    if isinstance(ref_value, Parameter):
                        raise TypeError(f"Part reference {ref_key} cannot be Parameter")
                    part.config[ref_name] = ref_value
                my_parts[platform_name].append((tmp, part.config))
        return my_parts

    def make_glues(self, platforms, agent_id: str, env_ref_stores: list[dict[str, typing.Any]]) -> None:
        """
        Creates agent glue functors from agent configuration.

        Parameters
        ----------
        platforms: The platform_name to platform instances mapping associated with the glue functors.
        agent_id: The id of the agent associated with the glue functors.
        env_ref_stores: Reference stores for items managed by the environment

        Returns
        -------
        None
        """
        sort_order = {}
        self.agent_glue_dict.clear()
        for config_glues, platform_name in zip(self.config.glues, self.config.platform_names):
            platform = platforms[platform_name]
            for glue_dict in config_glues:
                if self.config.normalization_master_switch == "ALL_OFF":
                    glue_dict.config["normalization"] = {"enabled": False}
                created_glue = glue_dict.create_functor_object(
                    platform=platform,
                    agent_name=agent_id,
                    param_sources=[self.local_variable_store.get("glues", {})],
                    ref_sources=[self.local_variable_store.get("reference_store", {}), self.config.reference_store, *env_ref_stores],
                )
                glue_name = created_glue.get_unique_name()
                if glue_name is None:
                    raise RuntimeError(f"No glue name for {created_glue}")

                if glue_name in self.agent_glue_dict:
                    raise RuntimeError(f"The {glue_name} glue has a unique name, but it already exists")
                # find dependent nodes for the topological search
                dependent_glues = {extractor.fields[0] for extractor in created_glue.config.extractors.values()}
                sort_order[glue_name] = dependent_glues

                self.agent_glue_dict[glue_name] = created_glue
        self.agent_glue_order = tuple(TopologicalSorter(sort_order).static_order())

    def make_rewards(self, agent_id: str, env_ref_stores: list[dict[str, typing.Any]]) -> None:
        """
        Creates agent reward functors from agent configuration.

        Parameters
        ----------
        agent_id: The id of the agent associated with the reward functors.
        env_ref_stores: Reference stores for items managed by the environment

        Returns
        -------
        None
        """
        tmp = [
            reward_dict.create_functor_object(
                agent_name=agent_id,
                platform_names=self.platform_names,
                param_sources=[self.local_variable_store.get("rewards", {})],
                ref_sources=[self.local_variable_store.get("reference_store", {}), self.config.reference_store, *env_ref_stores],
            )
            for reward_dict in self.config.rewards
        ]
        self.agent_reward_functors = tmp

    def make_platform_dones(
        self, env_params: dict[str, list[typing.Sequence[dict[str, typing.Any]]]], env_ref_stores: list[dict[str, typing.Any]]
    ) -> dict[str, list[DoneFuncBase]]:
        """
        Creates agent done functors from agent configuration.

        Parameters
        ----------
        agent_id: The id of the agent associated with the done functors.
        env_params: mapping of platform names to Parameters for the provided dones
        env_ref_stores: Reference stores for items managed by the environment

        Returns
        -------
        a dictionary containing mapping platform names to all dones the agent class wants
        to register to the platform
        """

        platform_dones: dict[str, list[DoneFuncBase]] = {}
        for platform_name in self.config.platform_names:
            my_platform_env_params = env_params[platform_name]
            platform_dones[platform_name] = []
            for done_dict in self.config.dones:
                platform_dones[platform_name].append(
                    done_dict.create_functor_object(
                        param_sources=[self.local_variable_store.get("dones", {}), *my_platform_env_params],  # type: ignore[list-item]
                        ref_sources=[self.local_variable_store.get("reference_store", {}), self.config.reference_store, *env_ref_stores],
                        agent_name=self.config.agent_name,
                        platform_name=platform_name,
                    )
                )
        return platform_dones

    def __create_space(
        self, return_space: dict | gymnasium.spaces.dict.Dict, space_getter: typing.Callable | None = None
    ) -> dict | gymnasium.spaces.dict.Dict | None:
        """
        Creates a gymnasium dict space from the agent's glues.

        Parameters
        ----------
        return_space typing.Union[typing.Dict, gymnasium.spaces.dict.Dict]: empty dict or space
        space_getter (optional): A function that takes a glue_obj: BaseAgentGlue and returns a space.
            Default is space_getter=lambda glue_obj: glue_obj.observation_space.

        Returns
        -------
        typing.Optional[typing.Union[typing.Dict, gymnasium.spaces.dict.Dict]]
            A gymnasium dict space or dict composed of all the spaces returned by applying the space_getter to the agent glues.
        """
        if space_getter is None:

            def default_getter(glue_obj):
                return glue_obj.observation_space

            space_getter = default_getter

        # loop over all glue name and  glue_obj pairs
        for glue_name in self.agent_glue_order:
            # call our space getter to pick which space we want,
            # for example: action_space, observation_space, normalized_action_space, normalized_observation_space
            space_def = space_getter(self.agent_glue_dict[glue_name])
            # if the space is None don't add it
            if space_def:
                return_space[glue_name] = space_def
        return return_space if len(return_space) > 0 else None

    def _create_space_dict(self, space_getter: typing.Callable | None = None):
        """
        Creates a dict from the agent's glues.

        Parameters
        ----------
        space_getter (optional): A function that takes a glue_obj: BaseAgentGlue and returns a space.
            Default is space_getter=lambda glue_obj: glue_obj.observation_space.

        Returns
        -------
        dict()
            A dict composed of all the spaces returned by applying the space_getter to the agent glues.
        """
        return self.__create_space({}, space_getter)

    def _create_space(self, space_getter: typing.Callable | None = None):
        """
        Creates a gymnasium dict space from the agent's glues.

        Parameters
        ----------
        space_getter (optional): A function that takes a glue_obj: BaseAgentGlue and returns a space.
            Default is space_getter=lambda glue_obj: glue_obj.observation_space.

        Returns
        -------
        gymnasium.spaces.dict.Dict()
            A gymnasium dict space composed of all the spaces returned by applying the space_getter to the agent glues.
        """
        return self.__create_space(gymnasium.spaces.dict.Dict(), space_getter)

    @cached_property
    def observation_prop(self) -> Prop:
        """
        Returns agents prop
        """
        return DictProp(spaces=self._create_space_dict(space_getter=lambda glue_obj: glue_obj.observation_prop))

    @cached_property
    def observation_units(self):
        """
        Returns agents normalized action space
        """
        return self.observation_prop.get_units()

    @cached_property
    def observation_space(self) -> gymnasium.Space:
        """
        Returns agents observation space
        """
        # calling it this way instead of getting it from the prop
        # causes the obs space cache to get hit on each glue instead of
        # recreating it from scratch using self.observation_prop.create_space()
        return self._create_space(space_getter=lambda glue_obj: glue_obj.observation_space)

    @cached_property
    def action_prop(self) -> Prop:
        """
        Returns the action prop that defines this agents actions
        """
        return DictProp(spaces=self._create_space_dict(space_getter=lambda glue_obj: glue_obj.action_prop))

    @cached_property
    def action_units(self) -> str | NestedDict:
        """
        Returns a nested dictionary where each leaf node is
        the unit of the observation matching the same access chain in
        the action
        """
        return self.action_prop.get_units()

    @cached_property
    def action_space(self) -> gymnasium.Space:
        """
        Returns agents action space
        """
        # calling it this way instead of getting it from the prop
        # causes the action space cache to get hit on each glue instead of
        # recreating it from scratch using self.action_prop.create_space()
        return self._create_space(space_getter=lambda glue_obj: glue_obj.action_space)

    @cached_property
    def normalized_observation_space(self) -> gymnasium.spaces.Space:
        """
        Returns agents normalized observation space
        """
        normalized_obs_space = self._create_space(
            space_getter=lambda glue_obj: glue_obj.normalized_observation_space
            if glue_obj.config.training_export_behavior == TrainingExportBehavior.INCLUDE
            else None
        )

        self._obs_space_transformation = self.__create_obs_space_transformation(normalized_obs_space)
        if self._obs_space_transformation:
            return self._obs_space_transformation.output_space
        return normalized_obs_space

    @cached_property
    def normalized_action_space(self):
        """
        Returns agents normalized action space
        """
        norm_space = self._create_space(space_getter=lambda glue_obj: glue_obj.normalized_action_space)

        self._action_space_transformation = self.__create_action_space_transformation(norm_space)

        if self._action_space_transformation:
            return self._action_space_transformation.output_space

        return norm_space

    def __create_action_space_transformation(self, space):
        """Create action space transformation"""
        if self.config.policy_action_transformation:
            return self.config.policy_action_transformation.transformation(space, self.config.policy_action_transformation.config)
        return None

    def __create_obs_space_transformation(self, space):
        """Create observations space transformation"""
        if self.config.policy_observation_transformation:
            return self.config.policy_observation_transformation.transformation(space, self.config.policy_observation_transformation.config)
        return None

    def apply_action(self, action: dict[typing.Any, typing.Any] | typing.Any, last_obs: typing.OrderedDict):
        """
        Applies actions to agent.

        Parameters
        ----------
        action (optional): Either a dictionary of actions to be applied to agent, or
            a flattened action.

        Returns
        -------
        None
        """

        if self._action_space_transformation:
            action = self._action_space_transformation.convert_transformed_sample(action)

        raw_action_dict = collections.OrderedDict()
        for glue_name, normalized_action in action.items():
            glue_object = self.agent_glue_dict[glue_name]
            raw_action = glue_object.unnormalize_action(normalized_action)
            raw_action = map_structure(lambda x, y: corl_get_ureg().Quantity(x, y), raw_action, self.action_units[glue_name])
            raw_action_dict[glue_name] = raw_action
            glue_object.apply_action(raw_action, last_obs, self.action_space, self.observation_space, self.observation_units)
        return raw_action_dict

    def create_next_action(self, observation: dict, reward: float, done: bool, info: dict):
        """
        Creates next action agent will apply.

        Parameters
        ----------
        action_dict (optional): A dictionary of actions to be applied to agent.

        Returns
        -------
        None
        """

    def get_observations(self) -> collections.OrderedDict:
        """
        Gets combined observation from agent glues.

        Returns
        -------
        OrderedDict
            A dictionary of glue observations in the form {glue_name: glue_observation}
        """
        return_observation: collections.OrderedDict = collections.OrderedDict()
        for glue_name in self.agent_glue_order:
            glue_object = self.agent_glue_dict[glue_name]
            if glue_obs := glue_object.get_observation(
                return_observation,
                self.observation_space,
                self.observation_units,
            ):
                return_observation[glue_name] = glue_obs
                if glue_object.config.clip_to_space and glue_object.observation_space:
                    return_observation[glue_name] = EnvSpaceUtil.clip_space_sample_to_space(glue_obs, glue_object.observation_space)

        return return_observation

    def get_info_dict(self) -> collections.OrderedDict:
        """
        Gets combined observation from agent glues.

        Returns
        -------
        OrderedDict
            A dictionary of glue observations in the form {glue_name: glue_observation}
        """
        return_info: collections.OrderedDict = collections.OrderedDict()
        for glue_name in self.agent_glue_order:
            if glue_info := self.agent_glue_dict[glue_name].get_info_dict():
                return_info[glue_name] = glue_info
        return return_info

    def get_rewards(self, observation, action, next_observation, state, next_state, observation_space, observation_units):
        """
        Get agent's environment rewards from agent reward functors.

        Parameters
        ----------
        - observation: The observation dictionary used to compute action.
        - action: The action computed from observation.
        - next_observation: The observation dictionary containing observation of next_state.
        - state: The state dictionary containing the environment state before action was applied to the environment.
        - next_state: The state dictionary containing the environment state after action was applied to the environment.
        - observation_space: The agent observation space.
        - observation_units: The units of the observations in the observation space. This may be None.

        Returns
        -------
        {agent_id: float}, {agent_id: Dict[str, float]}
        """
        reward_info = {}
        reward_total = 0.0
        for reward_functor in self.agent_reward_functors:
            output_value = reward_functor(
                observation=observation,
                action=action,
                next_observation=next_observation,
                state=state,
                next_state=next_state,
                observation_space=observation_space,
                observation_units=observation_units,
            )
            reward_total += output_value
            reward_info[reward_functor.name] = output_value

        return reward_total, reward_info

    def post_process_trajectory(self, agent_id, state, batch, episode, policy, reward_info):
        """
        calling function for sending data to on_postprocess_trajectory

        Arguments:
            agent_id: name of the agent
            state: current simulator state
            batch: batch from the current trajectory
            episode: the episode object
            policy: the policy that ran the trajectory
            reward_info: the info dict for rewards
        """
        for reward_functor in self.agent_reward_functors:
            output_value = reward_functor.post_process_trajectory(agent_id, state, batch, episode, policy)
            # If we have a bad setup the agent may be "dead" on first step
            # This in turn causes major issues when trying to index reward_info
            # Check that the agent exists
            if output_value is not None and agent_id in reward_info:
                reward_info[agent_id].setdefault(reward_functor.name, 0.0)
                reward_info[agent_id][reward_functor.name] += output_value

    def get_glue(self, name: str) -> BaseAgentGlue | None:
        """
        Get the glue object with the given name
        """
        return self.agent_glue_dict[name] if name in self.agent_glue_dict else None

    def normalize_observation(self, obs_name: str, obs: EnvSpaceUtil.sample_type) -> EnvSpaceUtil.sample_type | None:
        """
        Normalize a single observation value (if a corresponding glue is found)

        Parameters
        ----------
        - obs_name: The name of the observation
        - obs: the observation (value)

        Returns
        -------
        Normalized observation value, or None (if no corresponding glue is found)
        """
        glue_obj = self.get_glue(obs_name)
        return glue_obj.normalize_observation(obs) if glue_obj is not None else None

    def normalize_observations(self, observations: collections.OrderedDict) -> collections.OrderedDict:
        """
        Normalizes glue observations according to glue definition.

        Parameters
        ----------
        - observations: A dictionary of observation glues.

        Returns
        -------
        OrderedDict[str: typing.Union[np.ndarray, typing.Tuple, typing.Dict]]
            A dictionary of scaled observations in the form {glue_name: scaled_observation}
        """
        normalized_observation_dict = collections.OrderedDict()
        for obs_name, obs in observations:
            if normalized_obs := self.normalize_observation(obs_name, obs):
                normalized_observation_dict[obs_name] = normalized_obs

        return normalized_observation_dict

    def set_removed(self, removed_state: dict[str, bool]):
        """
        Set the agent_removed flag for the agent's glues.

        Parameters
        ----------
        - removed_state (typing.Dict[str, bool]): The inoperability status of the platform an agent class is controlling.
            True: the platform is no longer operable
            False: the platform is still operable

        Returns
        -------
        None
        """
        if not removed_state[self.platform_names[0]]:
            for glue in self.agent_glue_dict.values():
                glue.set_agent_removed(True)

    def create_training_observations(self, observations: collections.OrderedDict):
        """
        Base class representing a trainable agent in an environment.
        """

        training_obs = collections.OrderedDict()
        for obs_name, obs in observations.items():
            glue_obj = self.get_glue(obs_name)
            if glue_obj is not None and glue_obj.config.training_export_behavior == TrainingExportBehavior.INCLUDE:
                training_obs[obs_name] = self.normalize_observation(obs_name, obs)

        if self._obs_space_transformation:
            return self._obs_space_transformation.convert_sample(sample=training_obs)

        return training_obs

    def reset(self) -> None:
        """
        Method to allow child classes to call reset-dependent
        functionality, for instance at the end of an episode.
        """


class TrainableBaseAgent(BaseAgent):
    """
    Base class representing a trainable agent in an environment.
    """

    @property
    def trainable(self) -> bool:
        """
        Flag denoting if agent is trainable.

        Returns
        -------
        bool:
            True
        """
        return True


class NonTrainableBaseAgent(BaseAgent):
    """
    Base class representing a non-trainable agent in an environment.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def trainable(self) -> bool:
        """
        Flag denoting if agent is trainable.

        Returns
        -------
        bool:
            False
        """
        return False

    @abc.abstractmethod
    def create_action(self, observation: dict):
        """
        Creates an action given observation.

        Parameters
        ----------
        observation: dict
            agent's observation of the current environment

        Returns
        -------
        OrderedDict[str: Any]
            A dictionary of non-normalized actions applied to glues in
            the form {glue_name: raw_action}
        """



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Property Module
"""
import abc
import typing
from collections.abc import KeysView
from typing import TypeAlias

import gymnasium.spaces
import numpy as np
import tree
from gymnasium.spaces.utils import flatdim, unflatten
from numpy.typing import NDArray
from pydantic import BaseModel, ConfigDict, field_validator, model_validator
from ray.rllib.utils.spaces.repeated import Repeated

from corl.libraries.units import Quantity, corl_quantity

NestedDict: TypeAlias = dict[str, "str | NestedDict"]
NestedQuantity: TypeAlias = dict[str, "NestedQuantity | Quantity"]


class Prop(BaseModel, abc.ABC):
    """Represents the space prop outside of RLLIB"""

    name: str | None = None
    description: str | None = None
    model_config = ConfigDict(validate_default=True)

    def contains(self, sample) -> bool:
        """
        Creates RLLIB space
        """
        return self.create_space().contains(self.remove_units(sample))

    def sample(self) -> Quantity | NestedQuantity:
        """
        Creates RLLIB space
        """
        return self.create_quantity(self.create_space().sample())

    @abc.abstractmethod
    def get_units(self) -> str | NestedDict:
        """
        Gets the unit from this Prop
        """

    @abc.abstractmethod
    def create_space(self, seed=None) -> gymnasium.spaces.Space:
        """
        Creates gymnasium space based of property values
        """

    @abc.abstractmethod
    def create_unit_converted_prop(self, unit: str) -> "Prop":
        """
        This function is responsible for taking the current prop object and
        return, it may raise an exception if the prop is not capable of being a
        leaf node
        """

    @abc.abstractmethod
    def zero_mean(self) -> "Prop":
        """
        This function returns a version of this
        prop where all leaf nodes are scaled to be 0 mean
        """

    @abc.abstractmethod
    def scale(self, scale: float) -> "Prop":
        """
        This function returns a version of this
        prop where all leaf nodes are scaled by a scale factor
        """

    @abc.abstractmethod
    def create_quantity(self, value: dict | (float | (int | (list | np.ndarray)))) -> Quantity | NestedQuantity:
        """
        This function taskes in values and will attempt to create either a Quantity or NestedQuantity
        from it, properly applying units along the way
        """

    def remove_units(self, value: Quantity | NestedQuantity) -> dict | (float | (int | (list | np.ndarray))):  # noqa: PLR6301
        """
        This function takes in quanitites and will attempt to remove the units from the quantity
        """
        return tree.map_structure(lambda x: x.m, value)

    @abc.abstractmethod
    def create_zero_sample(self) -> Quantity | NestedQuantity:
        """
        This function will attempt to return 0 for each leaf node for all properties
        in a tree, as this is usually a safe default.  however if 0 is not in the low
        or high for this space it will return the lowest value possible
        Discrete values will always be the lowest, which should be 0
        """

    @abc.abstractmethod
    def create_low_sample(self) -> Quantity | NestedQuantity:
        """
        This function will attempt to return the lowest possible value for each leaf node
        for all properties in a tree.
        """


class BoxProp(Prop):
    """Represents the multi box outside of RLLIB"""

    dtype: np.dtype | None = np.dtype(np.float32)
    low: typing.Sequence[float] | typing.Sequence[typing.Sequence[float]] | np.ndarray
    high: typing.Sequence[float] | typing.Sequence[typing.Sequence[float]] | np.ndarray
    unit: str
    shape: tuple[int] | tuple[int, int] | None = None

    @field_validator("low", "high", mode="before")
    @classmethod
    def low_float_to_list(cls, v):
        return [v] if isinstance(v, int | float) else v

    @model_validator(mode="after")
    def prop_validator(cls, values):
        """Validate"""
        if not all(hasattr(values, k) for k in ("low", "high", "unit")):
            # Something went wrong, return
            return values

        # Broadcast space as needed
        space = gymnasium.spaces.Box(
            low=np.array(values.low).astype(values.dtype),
            high=np.array(values.high).astype(values.dtype),
            dtype=values.dtype,
            shape=values.shape,
        )
        values.low = space.low.tolist()
        values.high = space.high.tolist()
        values.shape = space.shape
        # Validate dimensions
        # 1D case
        assert len(values.low) == len(values.high), "low and high different length"
        if len(values.low) > 0 and isinstance(values.low[0], list):
            # list of lists 2D case
            for i, _ in enumerate(values.low):
                assert len(values.low[i]) == len(values.high[i]), "low and high list elements different length"

        return values

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_default=True)

    def create_space(self, seed=None) -> gymnasium.spaces.Space:
        """
        Creates RLLIB Box space
        """
        return gymnasium.spaces.Box(
            low=np.array(self.low).astype(self.dtype),
            high=np.array(self.high).astype(self.dtype),
            dtype=self.dtype,  # type: ignore
            shape=self.shape,
            seed=seed,
        )

    def min(self, convert: str | None = None) -> Quantity:  # noqa: A003
        """Min are pulled from sensor properties, converted to output units when necessary"""
        tmp = self.create_quantity(list(self.low))
        assert isinstance(tmp, Quantity)
        if convert:
            tmp = tmp.to(convert)
        return tmp

    def max(self, convert: str | None = None) -> Quantity:  # noqa: A003
        """Max are pulled from sensor properties, converted to output units when necessary"""
        tmp = self.create_quantity(list(self.high))
        assert isinstance(tmp, Quantity)
        if convert:
            tmp = tmp.to(convert)
        return tmp

    def create_converted_space(self, convert: str | None = None) -> gymnasium.spaces.Space:
        """
        Creates RLLIB Box space

        #TODO: seed??
        """
        return gymnasium.spaces.Box(
            low=np.array(self.min(convert).m).astype(self.dtype),
            high=np.array(self.max(convert).m).astype(self.dtype),
            dtype=self.dtype,  # type: ignore
            shape=self.shape,
        )

    def get_units(self):
        return self.unit

    def create_unit_converted_prop(self, unit: str):
        return type(self)(
            dtype=self.dtype,
            low=corl_quantity()(np.asarray(self.low), self.unit).to(unit).m.tolist(),
            high=corl_quantity()(np.asarray(self.high), self.unit).to(unit).m.tolist(),
            unit=unit,
            shape=self.shape,
        )

    def zero_mean(self) -> Prop:
        high = np.asarray(self.high)
        low = np.asarray(self.low)
        mean = (high + low) / 2
        return type(self)(low=(low - mean).tolist(), high=(high - mean).tolist(), dtype=self.dtype, unit=self.unit)

    def scale(self, scale: float) -> Prop:
        return type(self)(
            low=np.multiply(self.low, scale).astype(self.dtype).tolist(),
            high=np.multiply(self.high, scale).astype(self.dtype).tolist(),
            shape=self.shape,
            unit=self.unit,
            dtype=self.dtype,
        )

    def create_quantity(self, value: dict | float | int | list | np.ndarray) -> Quantity | NestedQuantity:
        if isinstance(value, float | int):
            new_value = np.array([value], dtype=self.dtype)
        elif isinstance(value, list):
            new_value = np.array(value, dtype=self.dtype)
        elif isinstance(value, np.ndarray):
            new_value = value.astype(self.dtype)
        else:
            raise RuntimeError(
                f"BoxProp tried to create a quantity from {value}, but BoxProp only knows how to make a quantity from "
                f" float, int, list, or np.ndarray, got {type(value)}"
            )
        new_value_shape = new_value.shape
        if self.shape != new_value_shape:
            raise RuntimeError(
                f"BoxProp tried to create a quantity from {value}, but the value given to create_quantity has a shape of {new_value_shape}"
                f" while this BoxProp is expecting to make shapes of {self.shape}"
            )
        return corl_quantity()(new_value, self.get_units())

    def create_zero_sample(self):
        new_value = np.asarray(self.low)
        new_value[new_value < 0.0] = 0.0
        assert np.all(new_value < np.asarray(self.high)), "0 sample ended up higher than high of space, use low_sample instead"
        return self.create_quantity(new_value)

    def create_low_sample(self):
        return self.min()


class DiscreteProp(Prop):
    """Represents the Discrete outside of RLLIB"""

    n: int

    def create_space(self, seed=None) -> gymnasium.spaces.Space:
        """
        Creates RLLIB Discrete space
        """
        return gymnasium.spaces.Discrete(self.n, seed=seed)

    def get_units(self):  # noqa: PLR6301
        return "dimensionless"

    def create_unit_converted_prop(self, unit: str):
        if unit != "dimensionless":
            raise RuntimeError(f"DiscreteProp was told to try and convert to unit {unit} but the only value unit is 'dimensionless'")
        return self

    def zero_mean(self) -> Prop:
        return type(self)(n=self.n)

    def scale(self, scale: float) -> Prop:
        return type(self)(n=self.n)

    def create_quantity(self, value):
        if not (isinstance(value, np.integer | int) or (isinstance(value, np.ndarray) and value.shape == ())):
            raise RuntimeError(f"Discrete Prop create_quantity can only handle integers, got {value=}, {type(value)}")
        return corl_quantity()(value, self.get_units())

    def create_zero_sample(self):
        return self.create_quantity(0)

    def create_low_sample(self):
        return self.create_zero_sample()


class MultiBinary(Prop):
    """Represents the multi binary outside of RLLIB"""

    n: int

    def create_space(self, seed=None) -> gymnasium.spaces.Space:
        """
        Creates RLLIB MultiBinary space
        """
        # TODO ray2 temporarily doesn't support multibinary correctly
        # so just map it to a vector of multi discretes
        return gymnasium.spaces.MultiDiscrete([2] * self.n, seed=seed, dtype=np.int8)

    def get_units(self):  # noqa: PLR6301
        return "dimensionless"

    def create_unit_converted_prop(self, unit: str):
        if unit != "dimensionless":
            raise RuntimeError(f"MultiBinary was told to try and convert to unit {unit} but the only value unit is 'dimensionless'")
        return self

    def zero_mean(self) -> Prop:
        return type(self)(n=self.n)

    def scale(self, scale: float) -> Prop:
        return type(self)(n=self.n)

    def create_quantity(self, value):
        assert isinstance(value, list | np.ndarray)
        return corl_quantity()(value, self.get_units())

    def create_zero_sample(self):
        return self.create_quantity([0] * self.n)

    def create_low_sample(self):
        return self.create_zero_sample()


class CorlRepeated(Repeated):  # noqa: PLW1641
    @property
    def is_np_flattenable(self):
        """Checks whether this space can be flattened to a :class:`spaces.Box`."""
        return self.child_space.is_np_flattenable

    def __eq__(self, other):
        return isinstance(other, Repeated) and self.child_space == other.child_space and self.max_len == other.max_len


@flatdim.register(CorlRepeated)
def _flatdim_dict(space: CorlRepeated) -> int:
    if space.is_np_flattenable:
        # length of repeated fields is the +1
        return flatdim(space.child_space) * space.max_len + 1
    raise ValueError(f"{space} cannot be flattened to a numpy array, probably because it contains a `Graph` or `Sequence` subspace")


@unflatten.register(CorlRepeated)
def _unflatten_corl_repeated(space: CorlRepeated, x: NDArray[typing.Any]) -> list[dict[str, typing.Any]]:
    # length of repeated fields is the +1
    entries = x[0]
    x = x[1:]
    if space.is_np_flattenable:
        dims = np.asarray([flatdim(space.child_space)] * space.max_len, dtype=np.int_)
        list_flattened = np.split(x, np.cumsum(dims[:-1]))
        return [unflatten(space.child_space, item) for item in list_flattened[0 : int(entries)]]
    raise ValueError(f"{space} cannot be unflattened to a {CorlRepeated}, probably because it contains a `Graph` or `Sequence` subspace")


class RepeatedProp(Prop):
    """Represents the multi binary outside of RLLIB"""

    max_len: int
    child_space: dict[str, Prop]

    def create_space(self, seed=None) -> gymnasium.spaces.Space:
        """
        Creates RLLIB Repeated space
        """
        gymnasium_child_space = gymnasium.spaces.Dict({key: value.create_space(seed=seed) for key, value in self.child_space.items()})
        return CorlRepeated(child_space=gymnasium_child_space, max_len=self.max_len)

    def get_units(self):
        return {key: value.get_units() for key, value in self.child_space.items()}

    def create_unit_converted_prop(self, unit: str):  # noqa: PLR6301
        raise RuntimeError(
            f"RepeatedProp was told to try and convert to unit {unit} but the only but a repeated Prop is not "
            f"capable of having a unit as it is not a leaf node.  this error should never happen and is just a sanity check"
            "to show this code path is not supported"
        )

    def zero_mean(self) -> Prop:
        return type(self)(
            child_space={field_name: field_prop.zero_mean() for field_name, field_prop in self.child_space.items()}, max_len=self.max_len
        )

    def scale(self, scale: float) -> Prop:
        return type(self)(
            child_space={field_name: field_prop.scale(scale) for field_name, field_prop in self.child_space.items()}, max_len=self.max_len
        )

    def create_quantity(self, value):
        assert isinstance(value, list)
        tmp_space = DictProp(spaces=self.child_space)
        new_ret = []
        for inner_row in value:
            assert isinstance(inner_row, dict)
            ret = tmp_space.create_quantity(inner_row)
            new_ret.append(ret)
        return new_ret

    def create_zero_sample(self):  # noqa: PLR6301
        return []

    def create_low_sample(self):  # noqa: PLR6301
        return []


class MultiDiscreteProp(Prop):
    """Represents the MultiDiscrete"""

    nvec: np.ndarray | list
    dtype: np.dtype = np.dtype(np.int32)
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def create_space(self, seed=None) -> gymnasium.spaces.Space:
        """
        Creates MultiDiscrete gymnasium space
        """
        return gymnasium.spaces.MultiDiscrete(self.nvec, dtype=self.dtype, seed=seed)  # type: ignore

    def get_units(self):  # noqa: PLR6301
        return "dimensionless"

    def create_unit_converted_prop(self, unit: str):
        if unit != "dimensionless":
            raise RuntimeError(f"MultiDiscreteProp was told to try and convert to unit {unit} but the only value unit is 'dimensionless'")
        return self

    def zero_mean(self) -> Prop:
        return type(self)(nvec=self.nvec, dtype=self.dtype)

    def scale(self, scale) -> Prop:
        return type(self)(nvec=self.nvec, dtype=self.dtype)

    def create_quantity(self, value):
        assert isinstance(value, list | np.ndarray)
        assert len(value) == len(self.nvec)
        return corl_quantity()(value, self.get_units())

    def create_zero_sample(self):
        return self.create_zero_sample([0] * len(self.nvec))

    def create_low_sample(self):
        return self.create_zero_sample()


class DictProp(Prop, typing.Mapping[str, Prop]):
    """Represents the MultiDiscrete"""

    spaces: dict[str, Prop]

    def create_space(self, seed=None) -> gymnasium.spaces.Dict:
        """
        Creates MultiDiscrete gymnasium space
        """
        return gymnasium.spaces.Dict(spaces={key: sub_prop.create_space(seed=seed) for key, sub_prop in self.items()})

    def get_units(self):
        ret = {}
        for key, prop in self.items():
            ret[key] = prop.get_units()
        return ret

    def create_unit_converted_prop(self, unit: str):  # noqa: PLR6301
        raise RuntimeError(
            f"DictProp was told to try and convert to unit {unit} but the only but a DictProp is not "
            f"capable of having a unit as it is not a leaf node.  this error should never happen and is just a sanity check"
            "to show this code path is not supported"
        )

    def zero_mean(self) -> Prop:
        return type(self)(spaces={field_name: field_prop.zero_mean() for field_name, field_prop in self.items()})

    def scale(self, scale) -> Prop:
        return type(self)(spaces={field_name: field_prop.scale(scale) for field_name, field_prop in self.items()})

    def create_quantity(self, value):
        assert isinstance(value, dict)
        ret = {}
        for prop_name, inner_prop in self.items():
            val = value.get(prop_name)
            if val is None:
                raise RuntimeError(
                    "Dict Prop tried to make quantity, but the value provided was not a valid value for this prop"
                    f"{value=}, space is {self.spaces}"
                )
            ret[prop_name] = inner_prop.create_quantity(val)
        return ret

    def create_zero_sample(self):
        ret = {}
        for prop_name, prop in self.items():
            ret[prop_name] = prop.create_zero_sample()

        return ret

    def create_low_sample(self):
        ret = {}
        for prop_name, prop in self.items():
            ret[prop_name] = prop.create_low_sample()

        return ret

    def __getitem__(self, key: str) -> Prop:
        """Get the space that is associated to `key`."""
        return self.spaces[key]

    def keys(self) -> KeysView:
        """Returns the keys of the Dict."""
        return KeysView(self.spaces)

    def __setitem__(self, key: str, value: Prop):
        """Set the space that is associated to `key`."""
        assert isinstance(value, Prop), f"Trying to set {key} to Dict prop with value that is not a Prop, actual type: {type(value)}"
        self.spaces[key] = value

    def __iter__(self):
        """Iterator through the keys of the subspaces."""
        yield from self.spaces

    def __len__(self) -> int:
        """Gives the number of simpler spaces that make up the `Dict` space."""
        return len(self.spaces)



"""
This module implements the Reward Functions and Reward Validators specific to the 1D Docking task.
"""

from collections import OrderedDict

import gymnasium
import numpy as np
from numpy_ringbuffer import RingBuffer

from corl.rewards.reward_func_base import RewardFuncBase, RewardFuncBaseValidator
from corl.simulators.base_simulator_state import BaseSimulatorState
from corl.simulators.common_platform_utils import get_platform_by_name, get_sensor_by_name


class DockingDistanceChangeRewardValidator(RewardFuncBaseValidator):
    """
    scale: Scalar value to adjust magnitude of the reward
    """

    scale: float = 1.0
    position_sensor_name: str


class DockingDistanceChangeReward(RewardFuncBase):
    """
    This RewardFuncBase extension is responsible for calculating the reward associated with a change in agent position.
    """

    def __init__(self, **kwargs) -> None:
        self.config: DockingDistanceChangeRewardValidator
        super().__init__(**kwargs)
        self._dist_buffer = RingBuffer(capacity=2, dtype=float)

    @staticmethod
    def get_validator() -> type[DockingDistanceChangeRewardValidator]:
        """
        Method to return class's Validator.
        """
        return DockingDistanceChangeRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: BaseSimulatorState,
        next_state: BaseSimulatorState,
        observation_space: gymnasium.Space,
        observation_units: OrderedDict,
    ) -> float:
        """
        This method calculates the current position of the agent and compares it to the previous position. The
        difference is used to return a proportional reward.

        Parameters
        ----------
        observation : OrderedDict
            The observations available to the agent from the previous state.
        action
            The last action performed by the agent.
        next_observation : OrderedDict
            The observations available to the agent from the current state.
        state : StateDict
            The previous state of the simulation.
        next_state : StateDict
            The current state of the simulation.
        observation_space : StateDict
            The agent's observation space.
        observation_units : StateDict
            The units corresponding to values in the observation_space?

        Returns
        -------
        reward
            The agent's reward for their change in distance.
        """
        val = 0

        deputy = get_platform_by_name(next_state, self.config.platform_names[0])
        position_sensor = get_sensor_by_name(deputy, self.config.position_sensor_name)  # type: ignore
        deputy_position = position_sensor.get_measurement()
        chief_position = np.array([0])  # hardcoded to origin

        distance = abs(chief_position - deputy_position.m)
        self._dist_buffer.append(distance[0])

        if len(self._dist_buffer) == 2:
            val = self.config.scale * ((self._dist_buffer[0] - self._dist_buffer[1]) < 0)

        return val



"""
This module defines functions that determine terminal conditions for the 1D Docking environment.
"""

from corl.dones.done_func_base import DoneFuncBase, DoneFuncBaseValidator, DoneStatusCodes
from corl.rewards.base_measurement_operation import ExtractorSet, ObservationExtractorValidator
from corl.simulators.common_platform_utils import get_platform_by_name
from corl.simulators.pong.pong import GameStatus


class CommanderPongGameDoneFunctionValidator(DoneFuncBaseValidator):
    """
    observation: observation extractor path to make sure a platform is still operable
    """

    observation: ObservationExtractorValidator


class CommanderPongGameDoneFunction(DoneFuncBase):
    """
    A done function that determines if the game status is no longer in progress and
    determines if the paddle this done is assigned to won or lost
    """

    def __init__(self, **kwargs) -> None:
        self.config: CommanderPongGameDoneFunctionValidator
        super().__init__(**kwargs)
        self.extractor: ExtractorSet = self.config.observation.construct_extractors()

    @staticmethod
    def get_validator() -> type[CommanderPongGameDoneFunctionValidator]:
        return CommanderPongGameDoneFunctionValidator

    def __call__(self, observation, action, next_observation, next_state, observation_space, observation_units):
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state

        Returns
        -------
        done : DoneDict
            dictionary containing the condition condition for the current agent

        """
        try:
            obs = self.extractor.value(observation).m.item()
        except:  # noqa: E722
            return False

        game_status = next_state.game_status

        done = False
        if game_status is not GameStatus.IN_PROGRESS:
            done = True
            if (game_status == GameStatus.LEFT_WIN and obs == 1) or (game_status == GameStatus.RIGHT_WIN and obs == 0):
                next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.WIN
            else:
                next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.LOSE

        return done


class CommanderPongHealthDone(DoneFuncBase):
    """
    A done function that determines if the current paddle has no health
    and needs to be set to done
    """

    def __call__(self, observation, action, next_observation, next_state, observation_space, observation_units):
        # Find Target Platform
        platform = get_platform_by_name(next_state, self.platform, allow_invalid=True)

        # platform does not exist
        if platform is None:
            return False

        return platform.paddle.current_health <= 0



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Loads pandas test cases
"""
import dataclasses
import enum
import logging
import typing

import jsonargparse
import pandas as pd
import yaml
from numpy.random import PCG64, Generator

from corl.evaluation.eval_logger_name import EVAL_LOGGER_NAME
from corl.evaluation.runners.section_factories.test_cases import config_parser, test_case_generator


@jsonargparse.typing.final
@dataclasses.dataclass
class Pandas:
    """Configure a Pandas Test Case"""

    class SourceForm(str, enum.Enum):
        """Format of the file given to this class"""

        FILE_CSV = "FILE_CSV"
        FILE_YAML_CONFIGURATION = "FILE_YAML_CONFIGURATION"
        BINARY = "BINARY"

    data: str | pd.DataFrame
    source_form: SourceForm

    data_frame: pd.DataFrame = dataclasses.field(init=False)

    seed: int = dataclasses.field(default=12345678903141592653589793)
    samples: float | None = dataclasses.field(default=None)
    randomize: bool = dataclasses.field(default=True)

    def __post_init__(self):
        self._name = None

    def name(self) -> str:
        """Get the name of the pandas test case

        Returns:
            str -- Name of pandas file
        """
        return self._name

    def generate(self):  # noqa: PLR0915
        """Generate pandas test cases"""

        logger = logging.getLogger(EVAL_LOGGER_NAME)

        test_cases = None
        randomize_config = None

        ###################################################
        ## Generate the test case data

        if self.source_form == Pandas.SourceForm.FILE_CSV:
            self._name = self.data.split("/")[-1]
            with open(self.data, encoding="utf=8") as f:
                test_cases_data = pd.read_csv(f, low_memory=False)
            logger.info("Reading test cases from %s", self.data)
            test_cases = test_cases_data
        elif self.source_form == Pandas.SourceForm.FILE_YAML_CONFIGURATION:
            self._name = self.data.split("/")[-1]
            with open(self.data, encoding="utf=8") as f:
                data = f.read()

            # Get a dict of the configuration for the randomized test cases
            config = yaml.safe_load(data)

            # Generate the value_config and the randomize_config
            if "episode_parameter_providers" in config:
                flat_config = config_parser.flatten_config(config.pop("episode_parameter_providers"))
                value_config = config_parser.resolve_config_ranges(
                    typing.cast(
                        config_parser.RESOLVED_VARIABLE_STRUCTURE_VALUE, config_parser.extract_config_element(flat_config, element="value")
                    )
                )
                randomize_config = typing.cast(
                    config_parser.EXTRACTED_VARIABLE_STRUCTURE_RANDOMIZE,
                    config_parser.extract_config_element(flat_config, element="randomize"),
                )
            else:
                value_config = {}
                randomize_config = {}

            # Read test cases
            test_cases = test_case_generator.create_test_cases(value_config)

        elif self.source_form == Pandas.SourceForm.BINARY:
            self._name = "evaluation"  # Not sure what to call it when given a binary
            test_cases = self.data
        else:
            raise RuntimeError(f"Unknown source form {self.source_form}")

        # Check that our parsing worked
        if test_cases is None:
            raise RuntimeError('"test_cases" variable is set to None, this is a non op as it should be set in the if/else tree above')

        ###############################################
        ## Now that we have test_cases loaded from _somewhere_ process them further

        # Not sure what this does, apparently needed
        if "test_case" in test_cases.columns:
            test_cases = test_cases.set_index("test_case")
        else:
            test_cases.index.name = "test_case"

        # Log
        logger.info("Number of test cases: %s", len(test_cases))
        logger.info("Number of variables: %s", len(test_cases.columns))

        # Some sanity checks
        if len(test_cases) == 0:
            raise ValueError("Initial test cases have length zero")

        # Configure random numbers
        logger.info("Random seed: %s", self.seed)
        rng = Generator(PCG64(self.seed))

        # Sample
        if self.samples is not None:
            logger.info("Sampling test cases")
            test_cases = test_case_generator.sample(rng.bit_generator, test_cases, self.samples)
            if len(test_cases) == 0:
                raise ValueError("Sampling reduced test cases to zero")
            logger.info("Tests cases remaining: %s", len(test_cases))

        # Randomize
        if self.randomize:
            if not randomize_config:
                raise ValueError("Cannot randomize without randomization parameters within the configuration")

            logger.info("Randomizing test cases")
            logger.info("Randomizing the following variables: %s", list(randomize_config.keys()))

            test_cases = test_case_generator.randomize_test_cases(test_cases=test_cases, gen=rng, config=randomize_config)

        self.data_frame = test_cases



"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
-------------
ObserveSensorRepeated Glue
"""
import typing
from collections import OrderedDict
from functools import cached_property

import gymnasium

from corl.glues.base_glue import BaseAgentPlatformGlue, BaseAgentPlatformGlueValidator
from corl.libraries.env_space_util import EnvSpaceUtil
from corl.libraries.property import DictProp, RepeatedProp
from corl.simulators.common_platform_utils import get_sensor_by_name


class ObserveSensorRepeatedValidator(BaseAgentPlatformGlueValidator):
    """
    sensor: which sensor to find on the platform
    output_units: unit to convert the output data to
    max_len: the maximum length to allow the observation space to reach
    """

    sensor: str
    output_units: dict[str, str] = {}
    max_len: int = 10


class ObserveSensorRepeated(BaseAgentPlatformGlue):
    """Glue to observe a sensor that can offer a variable length output

    Configuration:
        sensor: string      String of sensor class, sensor base name, or the PluginLibrary group name
                            For example consider a platform which has "SimOrientationRateSensor" attached.
                            The following will all be equivalent:
                            - SimOrientationRateSensor
                            - BaseOrientationRateSensor
                            - Sensor_OrientationRate

                            Important: If multiple sensors are found an array will be returned (ex. Gload)

                            This will also be the name attached to the glue, which can then be used
    """

    class Fields:
        """
        Fields in this glue
        """

        DIRECT_OBSERVATION = "direct_observation"

    @staticmethod
    def get_validator() -> type[ObserveSensorRepeatedValidator]:
        return ObserveSensorRepeatedValidator

    # TODO: self._sensor.measurement_properties assumes child_space attribute
    def __init__(self, **kwargs) -> None:
        self.config: ObserveSensorRepeatedValidator
        super().__init__(**kwargs)
        self._sensor = get_sensor_by_name(self._platform, self.config.sensor)
        self._sensor_name: str = self.config.sensor
        self.out_units = self.config.output_units
        self.max_len = self.config.max_len

    def get_unique_name(self) -> str:
        """Class method that retrieves the unique name for the glue instance"""
        # TODO: This may result in ObserveSensor_Sensor_X, which is kinda ugly
        return f"ObserveSensorRepeated_{self._sensor_name}"

    def invalid_value(self) -> OrderedDict:
        """Return zeros when invalid"""
        arr: list[float] = []
        d = OrderedDict()
        d[self.Fields.DIRECT_OBSERVATION] = arr
        return d

    @cached_property
    def observation_units(self):
        """Units of the sensors in this glue"""
        return self._observation_units

    @cached_property
    def observation_prop(self):
        tmp = self._sensor.measurement_properties
        assert isinstance(self._sensor.measurement_properties, RepeatedProp), "Unexpected measurement_properties type"
        child_space = tmp.child_space
        for prop_name, unit in self.out_units.items():
            child_space[prop_name] = child_space[prop_name].create_unit_converted_prop(unit)
        ret = RepeatedProp(child_space=child_space, max_len=self.max_len)
        return DictProp(spaces={self.Fields.DIRECT_OBSERVATION: ret})

    # TODO: Assumes s`elf._sensor.measurement_properties has child_space attribute
    def get_observation(self, other_obs: OrderedDict, obs_space, obs_units) -> OrderedDict:
        """Observation Values"""
        # NOTE: do not attempt to optimize this function by modifying data in place
        # you will mess up all other glues that call self._glue.get_observation
        # a copy is required here
        sensed_value = self._sensor.get_measurement()
        assert isinstance(sensed_value, list)
        tmp_sensed: list[dict[str, typing.Any]] = []
        append = tmp_sensed.append

        out_units = self.out_units
        for platform_data in sensed_value:
            tmp_row = {}
            for field_name, obs in platform_data.items():
                tmp_row[field_name] = obs.to(out_units[field_name]) if field_name in out_units else obs

            append(tmp_row)
            # we are going to clip the list to the max_len, so the obs is happy
            if len(tmp_sensed) == self.max_len:
                break

        d = OrderedDict()
        d[self.Fields.DIRECT_OBSERVATION] = tmp_sensed

        return d

    @cached_property
    def action_space(self) -> gymnasium.spaces.Space | None:
        """No Actions"""
        return None

    def apply_action(
        self,  # noqa: PLR6301
        action: EnvSpaceUtil.sample_type,
        observation: EnvSpaceUtil.sample_type,
        action_space,
        obs_space,
        obs_units,
    ):
        """No Actions"""
        return



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Entry / Launch point to run the evaluation framework processes, evaluation , generate_metrics , visualize back to back .
There are also various storage options
"""
from pathlib import Path
from typing import Any

import jsonargparse

from corl.evaluation.evaluation_artifacts import (
    EvaluationArtifact_EvaluationOutcome,
    EvaluationArtifact_Metrics,
    EvaluationArtifact_Visualization,
)
from corl.evaluation.launchers.base_eval import BaseProcessor, EvalConfig
from corl.evaluation.launchers.launch_evaluate import get_args as eval_get_args
from corl.evaluation.launchers.launch_evaluate import load_config as eval_load_config
from corl.evaluation.launchers.launch_evaluate import main as evaluate_main
from corl.evaluation.launchers.launch_generate_metrics import get_args as gen_metrics_get_args
from corl.evaluation.launchers.launch_generate_metrics import main as generate_metrics_main
from corl.evaluation.launchers.launch_storage import get_args as storage_get_args
from corl.evaluation.launchers.launch_storage import main as storage_main
from corl.evaluation.launchers.launch_visualize import get_args as visualize_get_args
from corl.evaluation.launchers.launch_visualize import main as visualize_main
from corl.evaluation.loader.policy_checkpoint import PolicyCheckpoint
from corl.evaluation.recording.folder import FolderRecord
from corl.evaluation.recording.i_recorder import IRecord
from corl.evaluation.runners.iterate_test_cases import ConfigSaver


class PipelineProcessor(BaseProcessor):
    def __init__(self, run_params: jsonargparse.Namespace, agent_checkpoints: dict[str, Path], **kwargs):
        super().__init__(**kwargs)
        self.run_params = run_params
        self.agent_checkpoints = agent_checkpoints

    def __call__(self, config: EvalConfig, results: Any):
        print("\n")
        print("Evaluation portion complete")
        use_eval_record = results[0]

        gen_metrics_params, metrics_artifact = gen_metrics_args(self.run_params, use_eval_record)

        print("\n")
        print("executing generate_metrics segment")

        generate_metrics_main(gen_metrics_params)

        print("\n")
        print("generate_metrics segment complete")

        _, visualize_params_instantiated, artifact_visualize = visualize_args(self.run_params, metrics_artifact)

        print("\n")
        print("executing visualization segment")

        visualize_main(visualize_params_instantiated)

        print("completed visualization segment")
        print("Evaluation framework pipeline complete")

        # collect visualization locations from visualize_params_instantiated

        if self.run_params.storage_config:
            artifact_location_config = {
                "eval_data_location": Path(str(use_eval_record.absolute_path)),
                "metrics_file_location": Path(str(metrics_artifact.location)).joinpath(str(metrics_artifact.file)),
                "visualizations_location": Path(str(artifact_visualize.location)),
                "agent_checkpoints": self.agent_checkpoints,
            }

            storage_instantiated_args = storage_args(self.run_params, artifact_location_config)
            storage_main(storage_instantiated_args)


def get_run_args() -> jsonargparse.Namespace:
    """
    Obtain the appropriate config arguments to run the evaluation framework pipeline.

    Returns
    -------
    run_params: jsonargparse.Namespace
        running configurations for the pipeline.
    """

    parser = jsonargparse.ArgumentParser()
    parser.add_argument("--cfg", action=jsonargparse.ActionConfigFile, help="path to yaml file containing run arguments")
    parser.add_argument("--eval_config", type=str, help="path to evaluation launch config")
    parser.add_argument("--gen_metrics_config", type=str, help="path to generate_metrics launch config")
    parser.add_argument("--visualize_config", type=str, help="path to visualize launch config")
    parser.add_argument("--storage_config", type=str, default=None, help="path to acedt bottle config")

    return parser.parse_args()


def get_evaluation_save(instantiated: type[jsonargparse.Namespace]) -> FolderRecord:
    """
    Obtain the location at which the evaluation segment recorded its results to. Picks up the first save location.

    Parameters:
        args: jsonargparse.Namespace
            contains the uninstantiated arguments from the evaluation segment.

    Returns:
        FolderRecord
            The first save location used in the evaluation segment to record evaluation results.
    """
    return instantiated["recorders"][0]  # type: ignore[index]


def get_agent_checkpoints(instantiated_args: jsonargparse.Namespace) -> dict[str, Path]:
    """
    From the evaluation segment retrieve the agent checkpoint locations.

    instantiated_args: jsonargparse.Namespace
        the instantiated classes arguments from the evaluation segment that setups the actual evaluation process

    Returns:
        agent_checkpoints: Dict[str,str]
            dictionary holding the agent checkpoints stored as agent_name to location.

    """
    agent_checkpoints = {}
    teams_map = instantiated_args["teams"]
    for agent in teams_map.agent_config:
        if isinstance(agent.agent_loader, PolicyCheckpoint):
            agent_checkpoints[agent.name] = Path(agent.agent_loader.checkpoint_filename)
    return agent_checkpoints


def evaluation_args(run_params: jsonargparse.Namespace) -> tuple[jsonargparse.Namespace, dict[str, Path]]:
    """
    Retrieve all run arguments to execute evaluation process and return the first save location for eval output.

    Parameters
    ----------
    run_params: jsonargparse.Namespace
        The running parameters for the pipeline

    Returns
    -------
    instantiated_args: jsonargparse.Namespace
        The instantiated classes needed to run the evaluation process
    eval_save_loc: str
        the first location to where the evaluation

    """

    cfg_file = run_params.eval_config

    instantiated_args, _ = eval_get_args(path=cfg_file)

    agent_checkpoints = get_agent_checkpoints(instantiated_args)

    return instantiated_args, agent_checkpoints


def gen_metrics_args(run_params: jsonargparse.Namespace, eval_record: IRecord) -> tuple[jsonargparse.Namespace, EvaluationArtifact_Metrics]:
    """
    Retrieve the running arguments for the generate_metrics process. Reconfigures locations as appropriate.

    Parameters
    ----------
    run_params: jsonargparse.Namespace
        configurations for the pipeline
    evaluation_outcome: EvaluationArtifact_EvaluationOutcome
        location of evaluation data output

    Returns
    -------
    args: jsonargparse.Namespace
        configurations for the gen_metrics process
    metrics_file_location:
        location of the metrics pkl file
    """

    gen_metrics_cfg = run_params.gen_metrics_config

    _, args = gen_metrics_get_args(gen_metrics_cfg)

    # by default grab the first record

    metrics_artifact = EvaluationArtifact_Metrics(str(eval_record.absolute_path))
    eval_artifact = EvaluationArtifact_EvaluationOutcome(location=eval_record)

    args["artifact_metrics"] = metrics_artifact
    args["artifact_evaluation_outcome"] = eval_artifact

    return args, metrics_artifact


def visualize_args(
    run_params: jsonargparse.Namespace, metrics_artifact: EvaluationArtifact_Metrics
) -> tuple[jsonargparse.Namespace, jsonargparse.Namespace, EvaluationArtifact_Visualization]:
    """
    Retrieve the running arguments for the visualize process. Reconfigures locations as appropriate.

    Parameters
    ----------
    run_params: jsonargparse.Namespace
        configurations for the pipeline
    metrics_artifact: EvaluationArtifact_Metrics
        Meta data of the generate metrics process
    Returns
    -------
    args: jsonargparse.Namespace
        configurations for the visualize process
    """

    visualize_cfg = run_params.visualize_config

    args, instantiate = visualize_get_args(visualize_cfg)

    common_folder = metrics_artifact.location
    visualization_folder = Path(common_folder).joinpath("visualizations")
    if not visualization_folder.is_dir():
        visualization_folder.mkdir()
    artifact_visualize = EvaluationArtifact_Visualization(str(visualization_folder))

    instantiate["artifact_metrics"] = metrics_artifact
    instantiate["artifact_visualization"] = artifact_visualize

    return args, instantiate, artifact_visualize


def storage_args(run_params, artifacts_config):
    """
    Retrieve the running arguments for the storage process. Reconfigures locations as appropriate.

    Parameters
    ----------
    run_params: jsonargparse.Namespace
        configurations for the pipeline
    artifacts_config: dict
        contains locations of all the artifacts from the evaluation framework pipeline.

    Returns
    -------
    instantiated: jsonargparse.Namespace
        configurations for the storage process
    """

    storage_instantiated = storage_get_args(run_params.storage_config)
    storage_instantiated["artifacts_location_config"] = artifacts_config
    return storage_instantiated


def main():
    """
    Run the evaluation framework pipeline. Evaluation -> Generate_metrics -> visualize -> storage (Optional)
    """

    run_params = get_run_args()

    args, agent_checkpoints = evaluation_args(run_params)

    print("Executing evaluation segment")

    config = eval_load_config(args.cfg)

    config["processors"] = [
        {"type": ConfigSaver},
        {"type": PipelineProcessor, "config": {"run_params": args, "agent_checkpoints": agent_checkpoints}},
    ]

    evaluate_main(args, config)


if __name__ == "__main__":
    main()



"""
-------------------------------------------------------------------------------
The Autonomous Capabilities Team (ACT3) Deep Reinforcement Learning (D-RL) Environment

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
-------------------------------------------------------------------------------

Common Platform Utils Module
"""
import typing

from corl.simulators.base_parts import BaseController, BasePlatformPart, BaseSensor
from corl.simulators.base_platform import BasePlatform
from corl.simulators.base_simulator import BaseSimulatorState


def get_platform_by_name(state: BaseSimulatorState, platform_name: str, allow_invalid=False) -> BasePlatform | None:
    """
    Gets a platform from a sim state based on a given agent id (name)

    Parameters
    ----------
    state: BaseSimulatorState
        State of current platforms in a sim step
    platform_name: str
        name of a given platform to search for in the sim state

    Returns
    -------
    platform: BasePlatform
        The platform with the given platform_name name
    """
    platform: BasePlatform | None = None
    platform = state.sim_platforms.get(platform_name, None)

    if not allow_invalid and (platform is None or not issubclass(platform.__class__, BasePlatform)):
        raise ValueError(f"Could not find a platform named {platform_name} of class BasePlatform")

    return platform


def get_controller_by_name(platform: BasePlatform, name: str) -> BaseController:
    """
    Gets a platform controller from a platform given the controller name

    Parameters
    ----------
    platform: BasePlatform
        platform to find part
    name: str
        name of a given controller to search for in the platform

    Returns
    -------
    platform_controller: BaseController
        The platform controller with the given name
    """
    return get_part_by_name(platform, name, BaseController)


def get_sensor_by_name(platform: BasePlatform, name: str) -> BaseSensor:
    """
    Gets a platform sensor from a platform given the sensor name

    Parameters
    ----------
    platform: BasePlatform
        platform to find part
    name: str
        name of a given sensor to search for in the platform

    Returns
    -------
    platform_sensor: BaseSensor
        The platform sensor with the given name
    """
    return get_part_by_name(platform, name, BaseSensor)


T = typing.TypeVar("T", bound=BasePlatformPart)


def get_part_by_name(platform: BasePlatform, name: str, part_type: type[T] | None = None) -> T:
    """
    Gets a platform part from a platform given the part name

    Parameters
    ----------
    platform: BasePlatform
        platform to find part
    name: str
        name of a given part  to search for in the platform

    Returns
    -------
    platform_part: BasePlatformPart
        The platform part with the given name
    """

    found_part = platform.controllers.get(name, None)
    if not found_part:
        found_part = platform.sensors.get(name, None)

    if not found_part:
        raise RuntimeError(f"An attached part associated with {name} group could not be found")
    if part_type and not isinstance(found_part, part_type):
        raise RuntimeError(
            f"An attached part associated with {name} group was found, however get_part_by_name "
            f"was requested to find a {part_type} and the found part {found_part} does not subclass that type"
        )

    return found_part


def is_platform_operable(state: BaseSimulatorState, platform_name: str) -> bool:
    """
    Check if a platform specified by name is operable

    Parameters
    ----------
    state:
        Simulation state
    platform_name:
        Name of the platform checking to be alive
    Returns
        Is platform operable?
    -------
    """

    platform = get_platform_by_name(state, platform_name, allow_invalid=True)
    return platform is not None and platform.operable



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import typing

from corl.evaluation.metrics.generator import MetricGeneratorAggregator
from corl.evaluation.metrics.metric import Metric, NonTerminalMetric
from corl.evaluation.metrics.types.nonterminals.timed_value import TimedValue
from corl.evaluation.metrics.types.nonterminals.vector import Vector
from corl.evaluation.metrics.types.terminals.void import Void


class Average(MetricGeneratorAggregator):
    """Aggregate by averaging. That is sum total and divide by number of occurrences"""

    def generate_metric(self, params: list[Metric] | Metric, **kwargs) -> Metric:  # noqa: PLR6301
        """Generate the metric

        Arguments:
            params {typing.Union[typing.List[Metric], Metric]} -- Already computed metrics

        Returns:
            Metric -- Computed Metric
        """

        arr: list[typing.Any]
        if isinstance(params, NonTerminalMetric):
            if isinstance(params, Vector):
                arr = params.arr
            else:
                raise RuntimeError(f"No known way to compute average of {type(params)}")

        elif isinstance(params, list):
            arr = params
        else:
            RuntimeError("Must either given a NonTerminalMetric or a list of terminals")

        total: Metric | None = None
        occurrence: int = 0
        for item in arr:
            if isinstance(item, Void):
                continue

            if isinstance(item, TimedValue):
                item = item.value  # noqa: PLW2901

            total = item if total is None else total + item
            occurrence += 1

        return Void() if total is None else total / occurrence



from corl.agents.base_agent import NonTrainableBaseAgent
from corl.glues.common.controller_glue import ControllerGlue


class RandomActionAgent(NonTrainableBaseAgent):
    """
    Agent that randomly samples control spce
    """

    def create_action(self, observation):
        """
        Creates a random action based on control space.

        Parameters
        ----------
        observation (object): agent's observation of the current environment

        Returns
        -------
        OrderedDict[str: Any]
        A dictionary of unnormalized actions applied to glues in the form {glue_name: raw_action}
        """
        control_dict = {}

        for glue_name, glue in self.agent_glue_dict.items():
            if isinstance(glue, ControllerGlue):
                control_dict[glue_name] = glue.action_space.sample()
        return control_dict



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

AvailablePlatforms
"""
import logging
import typing
from collections import OrderedDict
from functools import cached_property

import numpy as np
from gymnasium.spaces import Box, Dict
from pydantic import field_validator

from corl.glues.base_multi_wrapper import BaseMultiWrapperGlue, BaseMultiWrapperGlueValidator
from corl.glues.common.controller_glue import ControllerGlue
from corl.glues.common.observe_sensor import ObserveSensor
from corl.libraries.env_space_util import EnvSpaceUtil
from corl.libraries.property import BoxProp, DictProp
from corl.libraries.units import corl_get_ureg


class RelativeObsDeltaActionValidator(BaseMultiWrapperGlueValidator):
    """
    step_size:      A dict that contains a floating point scalar for each action in the action space,
                    by which the corresponding delta action is scaled prior to converting the action
                    to the wrapped space.
                    e.g. A throttle DeltaAction.apply_action([0.2]) with step_size=[.05] would move the
                    absolute throttle position to 0.01 higher than it was at the end of the last step.
    """

    step_size: float = 1.0
    obs_index: int | None = 0
    is_wrap: bool = False
    initial_value: float | None = None

    @field_validator("step_size")
    @classmethod
    def check_step_scale(cls, v):
        """
        verifies range of step scale values
        """
        if v >= 1.0 or v < 0:
            raise ValueError("RelativeObsDeltaActionValidator got step size of more that 1.0 or less than 0")
        return v


class RelativeObsDeltaAction(BaseMultiWrapperGlue):
    """
    RelativeObsDeltaAction is a glue class that wraps another glue class.
    It treats the actions passed to it as a delta from a linked observation
    E.G. if the wrapped action space has has roll as one of the controls, then a delta action of
    0.2 would move the absolute roll position 0.2 higher than it is as measured by the linked roll sensor.
    """

    def __init__(self, **kwargs) -> None:
        self.config: RelativeObsDeltaActionValidator
        super().__init__(**kwargs)

        self._logger = logging.getLogger(RelativeObsDeltaAction.__name__)

        if len(self.glues()) != 2:
            raise RuntimeError(f"Error: RelativeObsDeltaAction expected 2 wrapped glues, got {len(self.glues())}")
        self.controller: ControllerGlue = typing.cast(ControllerGlue, self.glues()[0])
        if not isinstance(self.controller, ControllerGlue):
            raise RuntimeError(
                f"Error: RelativeObsDeltaAction expects the first wrapped glue to be a ControllerGlue, got {self.controller}"
            )
        self.relative_obs_glue: ObserveSensor = typing.cast(ObserveSensor, self.glues()[1])
        if not isinstance(self.relative_obs_glue, ObserveSensor):
            raise RuntimeError(
                f"Error: RelativeObsDeltaAction expects the second wrapped glue to be a ObserveSensor, got {self.relative_obs_glue}"
            )

        # verify that the config setup is not going to get the user into a situation where they are
        # only accessing one part of the obs but applying that obs as the base position for multiple actions
        assert isinstance(self.controller.action_space, Dict)
        if self.config.obs_index and len(next(iter(self.controller.action_space.values())).low) != 1:  # type: ignore
            raise RuntimeError(
                f"ERROR: your glue {self.get_unique_name()} has an action space length of more than 1, "
                "but you specified though obs_index to access only 1 component of the obs "
                "from the wrapped observe Sensor, to fix this error in your config for this glue define 'obs_index': null"
            )

        self.step_size = EnvSpaceUtil.convert_config_param_to_space(
            action_space=self.controller.action_space, parameter=self.config.step_size
        )

        self._is_wrap = self.config.is_wrap

        self.saved_action_deltas = OrderedDict()
        assert isinstance(self.action_space, Dict)
        for space_name, space in self.action_space.items():
            if self.config.initial_value is not None:
                self.saved_action_deltas[space_name] = corl_get_ureg().Quantity(
                    np.array([0], dtype=self.action_prop.spaces[space_name].dtype), self.action_prop.spaces[space_name].get_units()
                )
            else:
                assert isinstance(space, Box)
                self.saved_action_deltas[space_name] = space.low

        wrapped_glue_name = self.controller.get_unique_name()
        if wrapped_glue_name is None:
            self._uname = None
        else:
            self._uname = f"{wrapped_glue_name}RelativeDelta"

    @staticmethod
    def get_validator() -> type[RelativeObsDeltaActionValidator]:
        return RelativeObsDeltaActionValidator

    def get_unique_name(self):
        """Class method that retrieves the unique name for the glue instance"""
        return self._uname

    def get_observation(self, other_obs: OrderedDict, obs_space, obs_units) -> np.ndarray | (tuple | dict):
        return {
            "absolute": self.controller.get_observation(other_obs, obs_space, obs_units),
            "delta": self.saved_action_deltas,
        }

    @cached_property
    def observation_prop(self):
        return DictProp(spaces={"absolute": self.controller.observation_prop, "delta": self.action_prop})

    @cached_property
    def action_prop(self):
        # # get the action space from the parent
        original_action_space = self.controller.action_prop

        # # zero mean the space so we can scale it easier
        zero_mean_space = original_action_space.zero_mean()

        # # scale the size of the unbiased space
        for space_name, space in zero_mean_space.spaces.items():
            zero_mean_space.spaces[space_name] = space.scale(scale=self.step_size[space_name])

        return zero_mean_space

    # TODO: assumes self.controller._control_properties has unit attribute
    def apply_action(self, action, observation, action_space, obs_space, obs_units) -> None:
        """
        Apply the action for the controllers
        """

        current_observation = self.relative_obs_glue.get_observation(observation, obs_space, obs_units)["direct_observation"]
        # all units in an array must be the same, so this assumption is ok
        obs_units = self.relative_obs_glue.observation_prop.get_units()["direct_observation"]
        if self.config.obs_index is not None:
            current_observation = current_observation[self.config.obs_index]
        assert isinstance(self.controller._control_properties, BoxProp), "Unexpected control_properties type"  # noqa: SLF001
        out_unit = self.controller._control_properties.unit  # noqa: SLF001
        assert isinstance(out_unit, str)
        unit_converted_obs = current_observation.to(out_unit)

        new_base_obs = OrderedDict()
        for control in action:
            new_base_obs[control] = unit_converted_obs

        self.saved_action_deltas = action

        assert self.action_space is not None
        absolute_action = EnvSpaceUtil.add_space_samples(
            space_template=self.action_space,
            space_sample1=action,
            space_sample2=new_base_obs,
        )
        assert self.controller.action_space is not None
        absolute_action = EnvSpaceUtil.clip_space_sample_to_space(absolute_action, self.controller.action_space, self._is_wrap)

        try:
            self.controller.apply_action(absolute_action, observation, action_space, obs_space, obs_units)
        except Exception as exc:  # noqa: BLE001
            # Purpose - add additional debugging information and re-raise the exception
            raise ValueError(
                f"\n"
                f"action={action}\n"
                f"current_observation={current_observation}\n"
                f"obs_unit={obs_units}\n"
                f"out_unit={out_unit}\n"
                f"action_space={self.action_space}\n"
                f"controller_action_space={self.controller.action_space}\n"
                f"is_wrap={self._is_wrap}\n"
            ) from exc

    def get_info_dict(self):  # noqa: PLR6301
        """
        Get the user specified metadata/metrics/etc.
        """
        return {}



"""
-------------------------------------------------------------------------------
The Autonomous Capabilities Team (ACT3) Deep Reinforcement Learning (D-RL) Environment

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
-------------------------------------------------------------------------------

Structures that hold parameters and the ability to update them.
"""
import abc
import typing
import warnings
from typing import runtime_checkable

import numpy as np
from numpy.random import Generator, RandomState
from pydantic import BaseModel, ConfigDict, PositiveFloat, StrictInt, TypeAdapter, field_validator, model_validator
from scipy import stats
from typing_extensions import Protocol

from corl.libraries.factory import Factory
from corl.libraries.units import Quantity, corl_get_ureg

Number = StrictInt | float
Randomness = Generator | RandomState
OtherVars = typing.Mapping[tuple[str, ...], Quantity]


@runtime_checkable
class _ConstraintCallbackType(Protocol):
    def __call__(self, old_arg: Number, new_arg: Number) -> Number:
        ...


class ParameterValidator(BaseModel):
    """Validator class for Parameter"""

    units: str = "dimensionless"
    update: dict[str, typing.Any] = {}
    simulator: dict[str, typing.Any] = {}
    episode_parameter_provider: dict[str, typing.Any] = {}
    dependent_parameters: dict[str, tuple[str, ...]] = {}

    @field_validator("dependent_parameters", mode="before")
    @classmethod
    def dependent_parameters_check(cls, v):
        """Validate the units field"""
        v2 = TypeAdapter(dict[str, str | typing.Sequence[str]]).validate_python(v)
        return {key: ("reference_store", item) if isinstance(item, str) else tuple(item) for key, item in v2.items()}


class Parameter(abc.ABC):
    """Parameter class"""

    def __init__(self, **kwargs) -> None:
        self.config: ParameterValidator = self.get_validator()(**kwargs)

        # Create and save updaters
        self.updaters: dict[str, typing.Any] = {}
        for name, val in self.config.update.items():
            factory = Factory(**val)
            self.updaters[name] = factory.build(param=self, name=name, constraint=self.get_constraint(name=name))

    @staticmethod
    def get_validator() -> type[ParameterValidator]:
        """Get the validator class for this Parameter"""
        return ParameterValidator

    def get_constraint(self, name: str) -> _ConstraintCallbackType | None:  # noqa: PLR6301
        """Get the constraint function for this Parameter's updater config"""
        return None

    @abc.abstractmethod
    def get_value(self, rng: Randomness, other_vars: OtherVars) -> str | Quantity:
        """Get the value of the parameter.

        In order to avoid inconsistent operation between cases where the value is serialized (such as from the ray object store) or not,
        this method should not modify the attributes of `self`.  This is the equivalent of a C++ `const` method for developers familiar with
        that concept.

        Parameters
        ----------
        rng : Union[Generator, RandomState]
            Random number generator from which to draw random values.
        other_vars: variables previously processed, will contain any variables this parameter is dependent on
        """

    # @staticmethod
    # def _serialize(v) -> str:
    #     return str(v)

    # @classmethod
    # def __get_pydantic_core_schema__(cls, source, handler):
    #     serializer = core_schema.plain_serializer_function_ser_schema(cls._serialize, when_used='json')
    #     if cls is source:
    #         # Treat bare usage of ImportString (`schema is None`) as the same as ImportString[Any]
    #         return core_schema.no_info_plain_validator_function(
    #             function=cls.validate, serialization=serializer
    #         )
    #     else:
    #         return core_schema.no_info_before_validator_function(
    #             function=cls.validate, schema=handler(source), serialization=serializer
    #         )

    # @classmethod
    # def validate(cls, v):
    #     if isinstance(v, Parameter):
    #         return v
    #     if not isinstance(v, dict):
    #         # msg = "Quantity validator input must either be a Quantity or Dict"
    #         # raise TypeError(msg)
    #         return v

    #     return tmp


class ParameterWrapperValidator(BaseModel):
    """Validator class for Parameter"""

    wrapped: dict[str, Parameter]
    # simulator: typing.Dict[str, typing.Any] = {}
    # episode_parameter_provider: typing.Dict[str, typing.Any] = {}
    dependent_parameters: dict[str, tuple[str, ...]] = {}
    model_config = ConfigDict(arbitrary_types_allowed=True)


class ParameterWrapper(abc.ABC):
    """ParameterWrapper class"""

    def __init__(self, **kwargs) -> None:
        self.config: ParameterWrapperValidator = self.get_validator()(**kwargs)

    @staticmethod
    def get_validator() -> type[ParameterWrapperValidator]:
        """Get the validator class for this Parameter"""
        return ParameterWrapperValidator

    def get_constraint(self, name: str) -> _ConstraintCallbackType | None:  # noqa: PLR6301
        """Get the constraint function for this Parameter's updater config"""
        return None

    @abc.abstractmethod
    def get_value(self, rng: Randomness, other_vars: OtherVars) -> dict[str, str | Quantity]:
        """Get the value of the parameters this Wrapper holds.

        This return object should contain a mapping to each parameter this wrapper wraps

        *IMPORTANT NOTE ON USE
        The environment will truncate the parameter tuple off the end and insert the string of the
        parameter key dict

        self.config.wrapped = {
            lat: ConstantParam(foo)
            lon: ConstantParam(bar)
        }

        ie: (group1, latlon (name of parameter)) -> (group1, lat) and (group2, lon)
        # END NOTE

        Parameters
        ----------
        rng : Union[Generator, RandomState]
            Random number generator from which to draw random values.
        other_vars: variables previously processed, will contain any variables this parameter is dependent on
        """


class PassthroughParameterWrapper(ParameterWrapper):
    """
    This parameter wrapper simply takes outputs the parameters it wraps
    into a dict
    """

    def get_value(self, rng, other_vars):
        return {k: v.get_value(rng, other_vars) for k, v in self.config.wrapped.items()}


class ConstantParameterValidator(ParameterValidator):
    """Validator class for ConstantParameter"""

    value: Number | str


class ConstantParameter(Parameter):
    """A parameter that always has a constant value."""

    def __init__(self, **kwargs) -> None:
        self.config: ConstantParameterValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[ConstantParameterValidator]:
        return ConstantParameterValidator

    def get_value(self, rng: Randomness, other_vars: OtherVars):
        if isinstance(self.config.value, str):
            return self.config.value
        return corl_get_ureg().Quantity(value=self.config.value, units=self.config.units)


class UniformParameterValidator(ParameterValidator):
    """Validator class for UniformParameter"""

    low: Number
    high: Number

    @model_validator(mode="after")
    def high_validator(self):
        """Validate the high field"""
        assert self.high > self.low, "Upper bound must not be smaller than lower bound"
        return self


class UniformParameter(Parameter):
    """A parameter that draws from a uniform distribution."""

    def __init__(self, **kwargs) -> None:
        self.config: UniformParameterValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[UniformParameterValidator]:
        return UniformParameterValidator

    def get_value(self, rng: Randomness, other_vars: OtherVars):
        return corl_get_ureg().Quantity(value=rng.uniform(self.config.low, self.config.high), units=self.config.units)

    def get_constraint(self, name: str) -> _ConstraintCallbackType | None:
        if name == "low":
            return self._min_with_high
        if name == "high":
            return self._max_with_low
        raise ValueError("Unknown constraint name")

    def _min_with_high(self, old_arg: Number, new_arg: Number) -> Number:
        if new_arg > self.config.high:
            warnings.warn("Could not fully update UniformParameter lower bound as it exceeds higher bound")
            return self.config.high
        return new_arg

    def _max_with_low(self, old_arg: Number, new_arg: Number) -> Number:
        if new_arg < self.config.low:
            warnings.warn("Could not fully update UniformParameter lower bound as it goes below the lower bound")
            return self.config.low
        return new_arg


class StepUniformParameterValidator(UniformParameterValidator):
    """Validator class for UniformParameter"""

    step: Number


class StepUniformParameter(Parameter):
    """A parameter that draws from a uniform distribution plus step."""

    def __init__(self, **kwargs) -> None:
        self.config: UniformParameterValidator
        super().__init__(**kwargs)
        if self.config.low == self.config.high:
            self._choices = [self.config.low]
        else:
            self._choices = list(range(self.config.low, self.config.high, self.config.step))  # type: ignore
        if self.config.high not in self._choices:
            self._choices.append(self.config.high)

    @staticmethod
    def get_validator() -> type[StepUniformParameterValidator]:
        return StepUniformParameterValidator

    def get_value(self, rng: Randomness, other_vars: OtherVars):
        return corl_get_ureg().Quantity(value=rng.choice(self._choices), units=self.config.units)


class TruncatedNormalParameterValidator(ParameterValidator):
    """Validator class for TruncatedNormalParameter"""

    mu: Number
    std: PositiveFloat
    half_width_factor: PositiveFloat


class TruncatedNormalParameter(Parameter):
    """A parameter that draws from a truncated normal distribution."""

    def __init__(self, **kwargs) -> None:
        self.config: TruncatedNormalParameterValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[TruncatedNormalParameterValidator]:
        return TruncatedNormalParameterValidator

    def get_value(self, rng: Randomness, other_vars: OtherVars) -> Quantity:
        low = self.config.mu - self.config.half_width_factor * self.config.std
        high = self.config.mu + self.config.half_width_factor * self.config.std
        value = stats.truncnorm.rvs(
            (low - self.config.mu) / self.config.std,
            (high - self.config.mu) / self.config.std,
            loc=self.config.mu,
            scale=self.config.std,
            size=1,
            random_state=rng,
        )[0]
        return corl_get_ureg().Quantity(value=value, units=self.config.units)

    def get_constraint(self, name: str) -> _ConstraintCallbackType | None:
        if name == "std":
            return self._std_positive
        if name == "half_width_factor":
            return self._half_width_factor_positive
        raise ValueError("Unknown constraint name")

    @staticmethod
    def _generic_positive(variable: str, old_arg: Number, new_arg: Number) -> Number:
        if new_arg < 0:
            warnings.warn(f"Could not update TruncatedNormalParameter {variable} because it is not strictly positive")
            return old_arg
        return new_arg

    def _std_positive(self, old_arg: Number, new_arg: Number) -> Number:
        return self._generic_positive(variable="standard deviation", old_arg=old_arg, new_arg=new_arg)

    def _half_width_factor_positive(self, old_arg: Number, new_arg: Number) -> Number:
        return self._generic_positive(variable="half width factor", old_arg=old_arg, new_arg=new_arg)


class ChoiceParameterValidator(ParameterValidator):
    """Validator for ChoiceParameter"""

    choices: typing.Sequence[typing.Any]


class ChoiceParameter(Parameter):
    """A parameter drawn uniformly from a collection of discrete values.
    This parameter does not support updaters.

    Parameters
    ----------
    hparams : dict
        The hyperparameters that define this parameter.  In addition to the structure specified by the base class Parameter, there needs
        to be the following fields, expressed as YAML:

        ```yaml
        choices: Sequence[Any]
        ```
    """

    def __init__(self, **kwargs) -> None:
        self.config: ChoiceParameterValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[ChoiceParameterValidator]:
        return ChoiceParameterValidator

    def get_value(self, rng: Randomness, other_vars: OtherVars):
        val = rng.choice(self.config.choices)
        if isinstance(val, str):
            return val
        return corl_get_ureg().Quantity(value=val, units=self.config.units)


class OverridableParameterWrapper(Parameter):
    """A Parameter that wraps another parameter and can override its output."""

    def __init__(self, base: Parameter | ParameterWrapper) -> None:
        # Base class API
        if isinstance(base, ParameterWrapper):
            raise RuntimeError(
                "OverridableParameterWrapper Does not support ParameterWrapper"
                "Evaluation test cases should be fully defined and should probably not use"
                "parameter wrappers"
            )
        self.config = base.config
        self.updaters = base.updaters

        # Other attributes
        self.base = base
        self.override_value: typing.Any = None

    @staticmethod
    def get_validator() -> typing.NoReturn:
        """OverridableParameterWrapper is not parsed using validators."""
        raise NotImplementedError()

    def get_constraint(self, name: str) -> _ConstraintCallbackType | None:
        return self.base.get_constraint(name=name)

    def get_value(self, rng: Randomness, other_vars: OtherVars):
        base_value = self.base.get_value(rng, other_vars)

        if self.override_value is not None:
            if isinstance(self.override_value, str):
                return self.override_value
            assert isinstance(base_value, Quantity)
            if isinstance(self.override_value, Quantity):
                return self.override_value.to(base_value.u)
            return corl_get_ureg().Quantity(value=self.override_value, units=base_value.u)

        return base_value


class UpdaterValidator(BaseModel):
    """Validator class for Updater"""

    name: str
    param: Parameter
    constraint: _ConstraintCallbackType = lambda old_arg, new_arg: new_arg
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("param")
    def param_hasattr(cls, v, info):
        """Validator for param field"""
        assert hasattr(v.config, info.data["name"]), f'Attribute {info.data["name"]} does not exist'
        return v

    @field_validator("constraint", mode="before")
    @classmethod
    def constraint_not_none(cls, v):
        """Conversion for constraint of None to identity"""
        return (lambda old_arg, new_arg: new_arg) if v is None else v


class Updater(abc.ABC):
    """Generic structure to define the method of updating a hyperparameter of a `Parameter`."""

    def __init__(self, **kwargs) -> None:
        # Validate and save updater configuration data
        self.config: UpdaterValidator = self.get_validator()(**kwargs)

    def __call__(self, *, reverse: bool = False) -> None:
        """Perform an update of the hyperparameter according to the functionality of the updater.

        The generic functionality of an update has four steps:
          1. Get the current value of the hyperparameter by extracting the named attribute from the connected parameter.
          2. Perform the update according to the `do_call` method within the updater.  The current value of the hyperparameter is provided,
             along with the flag to specify reverse updates.
          3. Apply the constraint function.  The primary purpose of this function is to allow the `Parameter` to provide a function to
             ensure that the updated hyperparameter value leaves the `Parameter` in a consistent state.  It receives the old argument and
             the new argument as keyword arguments (with names `old_arg` and `new_arg` respectively).  Common uses are to ensure that
             multiple hyperparameters within an `Parameter` maintain a consistent state (i.e., lower bound is less than upper bound) or that
             some invariant property is maintained (i.e., standard deviation is positive).  The constraint function must return the value
             that should be set into the hyperparameter.
          4. The output of the constraint function is set back to the named attribute in the connected parameter.

        Subclasses should not override `__call__`, but should implement `do_call`.

        Parameters
        ----------
        reverse : bool, optional
            Perform a reverse update, by default False
        """
        old_arg = self.get_current_extent()
        new_arg = self.do_call(old_arg, reverse=reverse)
        constrained_arg = self.config.constraint(old_arg=old_arg, new_arg=new_arg)
        setattr(self.config.param.config, self.config.name, constrained_arg)

    def update_to_bound(self) -> None:
        """Update all the way to the bound."""
        old_arg = self.get_current_extent()
        new_arg = self.get_bound()
        constrained_arg = self.config.constraint(old_arg=old_arg, new_arg=new_arg)
        setattr(self.config.param.config, self.config.name, constrained_arg)

    @abc.abstractmethod
    def do_call(self, arg: Number, *, reverse: bool = False) -> Number:
        """Perform the update of the provided hyperparameter.

        Parameters
        ----------
        arg : Union[int, float]
            The current value of the hyperparameter.
        reverse: bool, optional
            Perform a reverse update.  If allowed, performing a sequence with a normal update followed by a reverse update should return
            the original value.  Not all `Updater` subclasses allow reverse updates.  The default is False.
        """

    @staticmethod
    def get_validator() -> type[UpdaterValidator]:
        """Get the validator for this class"""
        return UpdaterValidator

    @abc.abstractmethod
    def supports_reverse_update(self) -> bool:
        """Indicator whether this updater supports reverse updates."""

    @abc.abstractmethod
    def at_bound(self) -> bool:
        """Indicator whether this updater is at its bound."""

    @abc.abstractmethod
    def get_bound(self) -> Number:
        """Get the bound of this updater."""

    def get_current_extent(self) -> Number:
        """returns the current extent of the parameter controlled by this updater

        Returns:
            Number -- The current extent of the parameter controlled by this updater
        """
        return getattr(self.config.param.config, self.config.name)

    @abc.abstractmethod
    def create_config(self) -> dict:
        """Create the configuration file that would generate this object in its current state.

        Returns
        -------
        dict
            Configuration file of the current state of the object.  This object can be passed to the
            constructor of the Parameter to regenerate it.
        """
        raise NotImplementedError()


class BoundStepUpdaterValidator(UpdaterValidator):
    """Validator class for BoundStepUpdater"""

    bound_type: typing.Literal["min", "max"]
    bound: Number
    step: Number

    @model_validator(mode="after")
    def step_validator(self):
        """Validator for step field"""
        if self.step >= 0 and self.bound_type == "min":
            raise ValueError("Step must be negative for minimum bound")
        if self.step <= 0 and self.bound_type == "max":
            raise ValueError("Step must be positive for maximum bound")
        return self


class BoundStepUpdater(Updater):
    """An `Updater` that advances by a constant step, limited by a bound.

    On each update, the provided value is incremented by a step size.  If that increment causes it to violate the provided bound, the value
    is given the value of that bound.

    Reverse updates are supported.  They are bounded at the initial value of the provided parameter.
    """

    def __init__(self, **kwargs) -> None:
        self.config: BoundStepUpdaterValidator
        super().__init__(**kwargs)

        self._reverse_bound = self.get_current_extent()

        self._bound_func: typing.Callable
        self._reverse_bound_func: typing.Callable
        if self.config.bound_type == "min":
            self._bound_func = max
            self._reverse_bound_func = min
        elif self.config.bound_type == "max":
            self._bound_func = min
            self._reverse_bound_func = max
        else:
            raise ValueError(f'Unknown bound type {self.config["bound_type"]}')
        self._at_bound: bool = False

    @staticmethod
    def get_validator() -> type[BoundStepUpdaterValidator]:
        """Get validator for BoundStepUpdater"""
        return BoundStepUpdaterValidator

    def do_call(self, arg: Number, *, reverse: bool = False) -> Number:
        if reverse:
            arg -= self.config.step
            output = self._reverse_bound_func(arg, self._reverse_bound)
        else:
            arg += self.config.step
            output = self._bound_func(arg, self.config.bound)

        self._at_bound = np.isclose(output, self.config.bound)

        return output

    def supports_reverse_update(self) -> bool:  # noqa: PLR6301
        return True

    def update_to_bound(self) -> None:
        super().update_to_bound()
        self._at_bound = True

    def at_bound(self) -> bool:
        return self._at_bound

    def get_bound(self) -> Number:
        return self.config.bound

    def create_config(self) -> dict:
        return {"bound": self.config.bound, "step": self.config.step, "bound_type": self.config.bound_type}


class RandomSignUniformParameter(UniformParameter):
    """A parameter that draws from a uniform distribution where the value is assigned a random sign.
    The range of the parameter is [-high, -low], [low, high]."""

    def get_value(self, rng: Randomness, other_vars: OtherVars):
        return corl_get_ureg().Quantity(value=rng.uniform(self.config.low, self.config.high) * rng.choice([-1, 1]), units=self.config.units)
